"""
Adapters for date/time types.
"""

# Copyright (C) 2020 The Psycopg Team

from __future__ import annotations

import re
import struct
from typing import TYPE_CHECKING, Any, cast
from datetime import date, datetime, time, timedelta, timezone
from collections.abc import Callable

from .. import _oids
from ..pq import Format
from .._tz import get_tzinfo
from ..abc import AdaptContext, DumperKey
from ..adapt import Buffer, Dumper, Loader, PyFormat
from ..errors import DataError, InterfaceError
from .._struct import pack_int4, pack_int8, unpack_int4, unpack_int8

if TYPE_CHECKING:
    from .._connection_base import BaseConnection

_struct_timetz = struct.Struct("!qi")  # microseconds, sec tz offset
_pack_timetz = cast(Callable[[int, int], bytes], _struct_timetz.pack)
_unpack_timetz = cast(Callable[[Buffer], "tuple[int, int]"], _struct_timetz.unpack)

_struct_interval = struct.Struct("!qii")  # microseconds, days, months
_pack_interval = cast(Callable[[int, int, int], bytes], _struct_interval.pack)
_unpack_interval = cast(
    Callable[[Buffer], "tuple[int, int, int]"], _struct_interval.unpack
)

utc = timezone.utc
_pg_date_epoch_days = date(2000, 1, 1).toordinal()
_pg_datetime_epoch = datetime(2000, 1, 1)
_pg_datetimetz_epoch = datetime(2000, 1, 1, tzinfo=utc)
_py_date_min_days = date.min.toordinal()


class DateDumper(Dumper):
    oid = _oids.DATE_OID

    def dump(self, obj: date) -> Buffer | None:
        # NOTE: whatever the PostgreSQL DateStyle input format (DMY, MDY, YMD)
        # the YYYY-MM-DD is always understood correctly.
        return str(obj).encode()


class DateBinaryDumper(Dumper):
    format = Format.BINARY
    oid = _oids.DATE_OID

    def dump(self, obj: date) -> Buffer | None:
        days = obj.toordinal() - _pg_date_epoch_days
        return pack_int4(days)


class _BaseTimeDumper(Dumper):
    def get_key(self, obj: time, format: PyFormat) -> DumperKey:
        # Use (cls,) to report the need to upgrade to a dumper for timetz (the
        # Frankenstein of the data types).
        if not obj.tzinfo:
            return self.cls
        else:
            return (self.cls,)

    def upgrade(self, obj: time, format: PyFormat) -> Dumper:
        raise NotImplementedError

    def _get_offset(self, obj: time) -> timedelta:
        if (offset := obj.utcoffset()) is None:
            raise DataError(
                f"cannot calculate the offset of tzinfo '{obj.tzinfo}' without a date"
            )
        return offset


class _BaseTimeTextDumper(_BaseTimeDumper):
    def dump(self, obj: time) -> Buffer | None:
        return str(obj).encode()


class TimeDumper(_BaseTimeTextDumper):
    oid = _oids.TIME_OID

    def upgrade(self, obj: time, format: PyFormat) -> Dumper:
        if not obj.tzinfo:
            return self
        else:
            return TimeTzDumper(self.cls)


class TimeTzDumper(_BaseTimeTextDumper):
    oid = _oids.TIMETZ_OID

    def dump(self, obj: time) -> Buffer | None:
        self._get_offset(obj)
        return super().dump(obj)


class TimeBinaryDumper(_BaseTimeDumper):
    format = Format.BINARY
    oid = _oids.TIME_OID

    def dump(self, obj: time) -> Buffer | None:
        us = obj.microsecond + 1_000_000 * (
            obj.second + 60 * (obj.minute + 60 * obj.hour)
        )
        return pack_int8(us)

    def upgrade(self, obj: time, format: PyFormat) -> Dumper:
        if not obj.tzinfo:
            return self
        else:
            return TimeTzBinaryDumper(self.cls)


class TimeTzBinaryDumper(_BaseTimeDumper):
    format = Format.BINARY
    oid = _oids.TIMETZ_OID

    def dump(self, obj: time) -> Buffer | None:
        us = obj.microsecond + 1_000_000 * (
            obj.second + 60 * (obj.minute + 60 * obj.hour)
        )
        off = self._get_offset(obj)
        return _pack_timetz(us, -int(off.total_seconds()))


class _BaseDatetimeDumper(Dumper):
    def get_key(self, obj: datetime, format: PyFormat) -> DumperKey:
        # Use (cls,) to report the need to upgrade (downgrade, actually) to a
        # dumper for naive timestamp.
        if obj.tzinfo:
            return self.cls
        else:
            return (self.cls,)

    def upgrade(self, obj: datetime, format: PyFormat) -> Dumper:
        raise NotImplementedError


class _BaseDatetimeTextDumper(_BaseDatetimeDumper):
    def dump(self, obj: datetime) -> Buffer | None:
        # NOTE: whatever the PostgreSQL DateStyle input format (DMY, MDY, YMD)
        # the YYYY-MM-DD is always understood correctly.
        return str(obj).encode()


class DatetimeDumper(_BaseDatetimeTextDumper):
    oid = _oids.TIMESTAMPTZ_OID

    def upgrade(self, obj: datetime, format: PyFormat) -> Dumper:
        if obj.tzinfo:
            return self
        else:
            return DatetimeNoTzDumper(self.cls)


class DatetimeNoTzDumper(_BaseDatetimeTextDumper):
    oid = _oids.TIMESTAMP_OID


class DatetimeBinaryDumper(_BaseDatetimeDumper):
    format = Format.BINARY
    oid = _oids.TIMESTAMPTZ_OID

    def dump(self, obj: datetime) -> Buffer | None:
        delta = obj - _pg_datetimetz_epoch
        micros = delta.microseconds + 1_000_000 * (86_400 * delta.days + delta.seconds)
        return pack_int8(micros)

    def upgrade(self, obj: datetime, format: PyFormat) -> Dumper:
        if obj.tzinfo:
            return self
        else:
            return DatetimeNoTzBinaryDumper(self.cls)


class DatetimeNoTzBinaryDumper(_BaseDatetimeDumper):
    format = Format.BINARY
    oid = _oids.TIMESTAMP_OID

    def dump(self, obj: datetime) -> Buffer | None:
        delta = obj - _pg_datetime_epoch
        micros = delta.microseconds + 1_000_000 * (86_400 * delta.days + delta.seconds)
        return pack_int8(micros)


class TimedeltaDumper(Dumper):
    oid = _oids.INTERVAL_OID

    def __init__(self, cls: type, context: AdaptContext | None = None):
        super().__init__(cls, context)
        if _get_intervalstyle(self.connection) == b"sql_standard":
            self._dump_method = self._dump_sql
        else:
            self._dump_method = self._dump_any

    def dump(self, obj: timedelta) -> Buffer | None:
        return self._dump_method(self, obj)

    @staticmethod
    def _dump_any(self: TimedeltaDumper, obj: timedelta) -> bytes:
        # The comma is parsed ok by PostgreSQL but it's not documented
        # and it seems brittle to rely on it. CRDB doesn't consume it well.
        return str(obj).encode().replace(b",", b"")

    @staticmethod
    def _dump_sql(self: TimedeltaDumper, obj: timedelta) -> bytes:
        # sql_standard format needs explicit signs
        # otherwise -1 day 1 sec will mean -1 sec
        return b"%+d day %+d second %+d microsecond" % (
            obj.days,
            obj.seconds,
            obj.microseconds,
        )


class TimedeltaBinaryDumper(Dumper):
    format = Format.BINARY
    oid = _oids.INTERVAL_OID

    def dump(self, obj: timedelta) -> Buffer | None:
        micros = 1_000_000 * obj.seconds + obj.microseconds
        return _pack_interval(micros, obj.days, 0)


class DateLoader(Loader):
    _ORDER_YMD = 0
    _ORDER_DMY = 1
    _ORDER_MDY = 2

    def __init__(self, oid: int, context: AdaptContext | None = None):
        super().__init__(oid, context)
        if (ds := _get_datestyle(self.connection)).startswith(b"I"):  # ISO
            self._order = self._ORDER_YMD
        elif ds.startswith(b"G"):  # German
            self._order = self._ORDER_DMY
        elif ds.startswith(b"S") or ds.startswith(b"P"):  # SQL or Postgres
            self._order = self._ORDER_DMY if ds.endswith(b"DMY") else self._ORDER_MDY
        else:
            raise InterfaceError(f"unexpected DateStyle: {ds.decode('ascii')}")

    def load(self, data: Buffer) -> date:
        if self._order == self._ORDER_YMD:
            ye = data[:4]
            mo = data[5:7]
            da = data[8:]
        elif self._order == self._ORDER_DMY:
            da = data[:2]
            mo = data[3:5]
            ye = data[6:]
        else:
            mo = data[:2]
            da = data[3:5]
            ye = data[6:]

        try:
            return date(int(ye), int(mo), int(da))
        except ValueError as ex:
            s = bytes(data).decode("utf8", "replace")
            if s == "infinity" or (s and len(s.split()[0]) > 10):
                raise DataError(f"date too large (after year 10K): {s!r}") from None
            elif s == "-infinity" or "BC" in s:
                raise DataError(f"date too small (before year 1): {s!r}") from None
            else:
                raise DataError(f"can't parse date {s!r}: {ex}") from None


class DateBinaryLoader(Loader):
    format = Format.BINARY

    def load(self, data: Buffer) -> date:
        days = unpack_int4(data)[0] + _pg_date_epoch_days
        try:
            return date.fromordinal(days)
        except (ValueError, OverflowError):
            if days < _py_date_min_days:
                raise DataError("date too small (before year 1)") from None
            else:
                raise DataError("date too large (after year 10K)") from None


class TimeLoader(Loader):
    _re_format = re.compile(rb"^(\d+):(\d+):(\d+)(?:\.(\d+))?")

    def load(self, data: Buffer) -> time:
        if not (m := self._re_format.match(data)):
            s = bytes(data).decode("utf8", "replace")
            raise DataError(f"can't parse time {s!r}")

        ho, mi, se, fr = m.groups()

        # Pad the fraction of second to get micros
        if fr:
            us = int(fr)
            if len(fr) < 6:
                us *= _uspad[len(fr)]
        else:
            us = 0

        try:
            return time(int(ho), int(mi), int(se), us)
        except ValueError as e:
            s = bytes(data).decode("utf8", "replace")
            raise DataError(f"can't parse time {s!r}: {e}") from None


class TimeBinaryLoader(Loader):
    format = Format.BINARY

    def load(self, data: Buffer) -> time:
        val = unpack_int8(data)[0]
        val, us = divmod(val, 1_000_000)
        val, s = divmod(val, 60)
        h, m = divmod(val, 60)
        try:
            return time(h, m, s, us)
        except ValueError:
            raise DataError(f"time not supported by Python: hour={h}") from None


class TimetzLoader(Loader):
    _re_format = re.compile(rb"""(?ix)
        ^
        (\d+) : (\d+) : (\d+) (?: \. (\d+) )?       # Time and micros
        ([-+]) (\d+) (?: : (\d+) )? (?: : (\d+) )?  # Timezone
        $
        """)

    def load(self, data: Buffer) -> time:
        if not (m := self._re_format.match(data)):
            s = bytes(data).decode("utf8", "replace")
            raise DataError(f"can't parse timetz {s!r}")

        ho, mi, se, fr, sgn, oh, om, os = m.groups()

        # Pad the fraction of second to get the micros
        if fr:
            us = int(fr)
            if len(fr) < 6:
                us *= _uspad[len(fr)]
        else:
            us = 0

        # Calculate timezone
        off = 60 * 60 * int(oh)
        if om:
            off += 60 * int(om)
        if os:
            off += int(os)
        tz = timezone(timedelta(0, off if sgn == b"+" else -off))

        try:
            return time(int(ho), int(mi), int(se), us, tz)
        except ValueError as e:
            s = bytes(data).decode("utf8", "replace")
            raise DataError(f"can't parse timetz {s!r}: {e}") from None


class TimetzBinaryLoader(Loader):
    format = Format.BINARY

    def load(self, data: Buffer) -> time:
        val, off = _unpack_timetz(data)

        val, us = divmod(val, 1_000_000)
        val, s = divmod(val, 60)
        h, m = divmod(val, 60)

        try:
            return time(h, m, s, us, timezone(timedelta(seconds=-off)))
        except ValueError:
            raise DataError(f"time not supported by Python: hour={h}") from None


class TimestampLoader(Loader):
    _re_format = re.compile(rb"""(?ix)
        ^
        (\d+) [^a-z0-9] (\d+) [^a-z0-9] (\d+)   # Date
        (?: T | [^a-z0-9] )                     # Separator, including T
        (\d+) [^a-z0-9] (\d+) [^a-z0-9] (\d+)   # Time
        (?: \.(\d+) )?                          # Micros
        $
        """)
    _re_format_pg = re.compile(rb"""(?ix)
        ^
        [a-z]+          [^a-z0-9]               # DoW, separator
        (\d+|[a-z]+)    [^a-z0-9]               # Month or day
        (\d+|[a-z]+)    [^a-z0-9]               # Month or day
        (\d+) [^a-z0-9] (\d+) [^a-z0-9] (\d+)   # Time
        (?: \.(\d+) )?                          # Micros
        [^a-z0-9] (\d+)                         # Year
        $
        """)

    _ORDER_YMD = 0
    _ORDER_DMY = 1
    _ORDER_MDY = 2
    _ORDER_PGDM = 3
    _ORDER_PGMD = 4

    def __init__(self, oid: int, context: AdaptContext | None = None):
        super().__init__(oid, context)
        if (ds := _get_datestyle(self.connection)).startswith(b"I"):  # ISO
            self._order = self._ORDER_YMD
        elif ds.startswith(b"G"):  # German
            self._order = self._ORDER_DMY
        elif ds.startswith(b"S"):  # SQL
            self._order = self._ORDER_DMY if ds.endswith(b"DMY") else self._ORDER_MDY
        elif ds.startswith(b"P"):  # Postgres
            self._order = self._ORDER_PGDM if ds.endswith(b"DMY") else self._ORDER_PGMD
            self._re_format = self._re_format_pg
        else:
            raise InterfaceError(f"unexpected DateStyle: {ds.decode('ascii')}")

    def load(self, data: Buffer) -> datetime:
        if not (m := self._re_format.match(data)):
            raise _get_timestamp_load_error(self.connection, data) from None

        if self._order == self._ORDER_YMD:
            ye, mo, da, ho, mi, se, fr = m.groups()
            imo = int(mo)
        elif self._order == self._ORDER_DMY:
            da, mo, ye, ho, mi, se, fr = m.groups()
            imo = int(mo)
        elif self._order == self._ORDER_MDY:
            mo, da, ye, ho, mi, se, fr = m.groups()
            imo = int(mo)
        else:
            if self._order == self._ORDER_PGDM:
                da, mo, ho, mi, se, fr, ye = m.groups()
            else:
                mo, da, ho, mi, se, fr, ye = m.groups()
            try:
                imo = _month_abbr[mo]
            except KeyError:
                s = mo.decode("utf8", "replace")
                raise DataError(f"can't parse month: {s!r}") from None

        # Pad the fraction of second to get the micros
        if fr:
            us = int(fr)
            if len(fr) < 6:
                us *= _uspad[len(fr)]
        else:
            us = 0

        try:
            return datetime(int(ye), imo, int(da), int(ho), int(mi), int(se), us)
        except ValueError as ex:
            raise _get_timestamp_load_error(self.connection, data, ex) from None


class TimestampBinaryLoader(Loader):
    format = Format.BINARY

    def load(self, data: Buffer) -> datetime:
        micros = unpack_int8(data)[0]
        try:
            return _pg_datetime_epoch + timedelta(microseconds=micros)
        except OverflowError:
            if micros <= 0:
                raise DataError("timestamp too small (before year 1)") from None
            else:
                raise DataError("timestamp too large (after year 10K)") from None


class TimestamptzLoader(Loader):
    _re_format = re.compile(rb"""(?ix)
        ^
        (\d+) [^a-z0-9] (\d+) [^a-z0-9] (\d+)       # Date
        (?: T | [^a-z0-9] )                         # Separator, including T
        (\d+) [^a-z0-9] (\d+) [^a-z0-9] (\d+)       # Time
        (?: \.(\d+) )?                              # Micros
        ([-+]) (\d+) (?: : (\d+) )? (?: : (\d+) )?  # Timezone
        $
        """)

    def __init__(self, oid: int, context: AdaptContext | None = None):
        super().__init__(oid, context)
        self._timezone = get_tzinfo(self.connection.pgconn if self.connection else None)

        if _get_datestyle(self.connection).startswith(b"I"):  # ISO
            self._load_method = self._load_iso
        else:
            self._load_method = self._load_notimpl

    def load(self, data: Buffer) -> datetime:
        return self._load_method(self, data)

    @staticmethod
    def _load_iso(self: TimestamptzLoader, data: Buffer) -> datetime:
        if not (m := self._re_format.match(data)):
            raise _get_timestamp_load_error(self.connection, data) from None

        ye, mo, da, ho, mi, se, fr, sgn, oh, om, os = m.groups()

        # Pad the fraction of second to get the micros
        if fr:
            us = int(fr)
            if len(fr) < 6:
                us *= _uspad[len(fr)]
        else:
            us = 0

        # Calculate timezone offset
        soff = 60 * 60 * int(oh)
        if om:
            soff += 60 * int(om)
        if os:
            soff += int(os)
        tzoff = timedelta(0, soff if sgn == b"+" else -soff)

        # The return value is a datetime with the timezone of the connection
        # (in order to be consistent with the binary loader, which is the only
        # thing it can return). So create a temporary datetime object, in utc,
        # shift it by the offset parsed from the timestamp, and then move it to
        # the connection timezone.
        dt = None
        ex: Exception
        try:
            dt = datetime(int(ye), int(mo), int(da), int(ho), int(mi), int(se), us, utc)
            return (dt - tzoff).astimezone(self._timezone)
        except OverflowError as e:
            # If we have created the temporary 'dt' it means that we have a
            # datetime close to max, the shift pushed it past max, overflowing.
            # In this case return the datetime in a fixed offset timezone.
            if dt is not None:
                return dt.replace(tzinfo=timezone(tzoff))
            else:
                ex = e
        except ValueError as e:
            ex = e

        raise _get_timestamp_load_error(self.connection, data, ex) from None

    @staticmethod
    def _load_notimpl(self: TimestamptzLoader, data: Buffer) -> datetime:
        s = bytes(data).decode("utf8", "replace")
        ds = _get_datestyle(self.connection).decode("ascii")
        raise NotImplementedError(
            f"can't parse timestamptz with DateStyle {ds!r}: {s!r}"
        )


class TimestamptzBinaryLoader(Loader):
    format = Format.BINARY

    def __init__(self, oid: int, context: AdaptContext | None = None):
        super().__init__(oid, context)
        self._timezone = get_tzinfo(self.connection.pgconn if self.connection else None)

    def load(self, data: Buffer) -> datetime:
        micros = unpack_int8(data)[0]
        try:
            ts = _pg_datetimetz_epoch + timedelta(microseconds=micros)
            return ts.astimezone(self._timezone)
        except OverflowError:
            # If we were asked about a timestamp which would overflow in UTC,
            # but not in the desired timezone (e.g. datetime.max at Chicago
            # timezone) we can still save the day by shifting the value by the
            # timezone offset and then replacing the timezone.
            if self._timezone:
                if utcoff := self._timezone.utcoffset(
                    datetime.min if micros < 0 else datetime.max
                ):
                    usoff = 1_000_000 * int(utcoff.total_seconds())
                    try:
                        ts = _pg_datetime_epoch + timedelta(microseconds=micros + usoff)
                    except OverflowError:
                        pass  # will raise downstream
                    else:
                        return ts.replace(tzinfo=self._timezone)

            if micros <= 0:
                raise DataError("timestamp too small (before year 1)") from None
            else:
                raise DataError("timestamp too large (after year 10K)") from None


class IntervalLoader(Loader):
    _re_interval = re.compile(
        rb"""
        (?: ([-+]?\d+) \s+ years? \s* )?                # Years
        (?: ([-+]?\d+) \s+ mons? \s* )?                 # Months
        (?: ([-+]?\d+) \s+ days? \s* )?                 # Days
        (?: ([-+])? (\d+) : (\d+) : (\d+ (?:\.\d+)?)    # Time
        )?
        """,
        re.VERBOSE,
    )

    def __init__(self, oid: int, context: AdaptContext | None = None):
        super().__init__(oid, context)
        if _get_intervalstyle(self.connection) == b"postgres":
            self._load_method = self._load_postgres
        else:
            self._load_method = self._load_notimpl

    def load(self, data: Buffer) -> timedelta:
        return self._load_method(self, data)

    @staticmethod
    def _load_postgres(self: IntervalLoader, data: Buffer) -> timedelta:
        if not (m := self._re_interval.match(data)):
            s = bytes(data).decode("utf8", "replace")
            raise DataError(f"can't parse interval {s!r}")

        ye, mo, da, sgn, ho, mi, se = m.groups()
        days = 0
        seconds = 0.0

        if ye:
            days += 365 * int(ye)
        if mo:
            days += 30 * int(mo)
        if da:
            days += int(da)

        if ho:
            seconds = 3600 * int(ho) + 60 * int(mi) + float(se)
            if sgn == b"-":
                seconds = -seconds

        try:
            return timedelta(days=days, seconds=seconds)
        except OverflowError as e:
            s = bytes(data).decode("utf8", "replace")
            raise DataError(f"can't parse interval {s!r}: {e}") from None

    @staticmethod
    def _load_notimpl(self: IntervalLoader, data: Buffer) -> timedelta:
        s = bytes(data).decode("utf8", "replace")
        ints = _get_intervalstyle(self.connection).decode("utf8", "replace")
        raise NotImplementedError(
            f"can't parse interval with IntervalStyle {ints}: {s!r}"
        )


class IntervalBinaryLoader(Loader):
    format = Format.BINARY

    def load(self, data: Buffer) -> timedelta:
        micros, days, months = _unpack_interval(data)
        if months > 0:
            years, months = divmod(months, 12)
            days = days + 30 * months + 365 * years
        elif months < 0:
            years, months = divmod(-months, 12)
            days = days - 30 * months - 365 * years

        try:
            return timedelta(days=days, microseconds=micros)
        except OverflowError as e:
            raise DataError(f"can't parse interval: {e}") from None


def _get_datestyle(conn: BaseConnection[Any] | None) -> bytes:
    if conn and (ds := conn.pgconn.parameter_status(b"DateStyle")):
        return ds

    return b"ISO, DMY"


def _get_intervalstyle(conn: BaseConnection[Any] | None) -> bytes:
    if conn and (ints := conn.pgconn.parameter_status(b"IntervalStyle")):
        return ints

    return b"unknown"


def _get_timestamp_load_error(
    conn: BaseConnection[Any] | None, data: Buffer, ex: Exception | None = None
) -> Exception:
    s = bytes(data).decode("utf8", "replace")

    def is_overflow(s: str) -> bool:
        if not s:
            return False

        if not _get_datestyle(conn).startswith(b"P"):  # Postgres
            return len(s.split()[0]) > 10  # date is first token
        else:
            return len(s.split()[-1]) > 4  # year is last token

    if s == "-infinity" or s.endswith("BC"):
        return DataError("timestamp too small (before year 1): {s!r}")
    elif s == "infinity" or is_overflow(s):
        return DataError(f"timestamp too large (after year 10K): {s!r}")
    else:
        return DataError(f"can't parse timestamp {s!r}: {ex or '(unknown)'}")


_month_abbr = {
    n: i
    for i, n in enumerate(b"Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec".split(), 1)
}

# Pad to get microseconds from a fraction of seconds
_uspad = [0, 100_000, 10_000, 1_000, 100, 10, 1]


def register_default_adapters(context: AdaptContext) -> None:
    adapters = context.adapters
    adapters.register_dumper("datetime.date", DateDumper)
    adapters.register_dumper("datetime.date", DateBinaryDumper)

    # first register dumpers for 'timetz' oid, then the proper ones on time type.
    adapters.register_dumper("datetime.time", TimeTzDumper)
    adapters.register_dumper("datetime.time", TimeTzBinaryDumper)
    adapters.register_dumper("datetime.time", TimeDumper)
    adapters.register_dumper("datetime.time", TimeBinaryDumper)

    # first register dumpers for 'timestamp' oid, then the proper ones
    # on the datetime type.
    adapters.register_dumper("datetime.datetime", DatetimeNoTzDumper)
    adapters.register_dumper("datetime.datetime", DatetimeNoTzBinaryDumper)
    adapters.register_dumper("datetime.datetime", DatetimeDumper)
    adapters.register_dumper("datetime.datetime", DatetimeBinaryDumper)

    adapters.register_dumper("datetime.timedelta", TimedeltaDumper)
    adapters.register_dumper("datetime.timedelta", TimedeltaBinaryDumper)

    adapters.register_loader("date", DateLoader)
    adapters.register_loader("date", DateBinaryLoader)
    adapters.register_loader("time", TimeLoader)
    adapters.register_loader("time", TimeBinaryLoader)
    adapters.register_loader("timetz", TimetzLoader)
    adapters.register_loader("timetz", TimetzBinaryLoader)
    adapters.register_loader("timestamp", TimestampLoader)
    adapters.register_loader("timestamp", TimestampBinaryLoader)
    adapters.register_loader("timestamptz", TimestamptzLoader)
    adapters.register_loader("timestamptz", TimestamptzBinaryLoader)
    adapters.register_loader("interval", IntervalLoader)
    adapters.register_loader("interval", IntervalBinaryLoader)



from enum import Enum, auto

import pytest

from psycopg import ClientCursor
from psycopg import errors as e
from psycopg import pq, sql
from psycopg.adapt import PyFormat
from psycopg.types import TypeInfo
from psycopg.types.enum import EnumInfo, register_enum

from ..fix_crdb import crdb_encoding


class PureTestEnum(Enum):
    FOO = auto()
    BAR = auto()
    BAZ = auto()


class StrTestEnum(str, Enum):
    ONE = "ONE"
    TWO = "TWO"
    THREE = "THREE"


NonAsciiEnum = Enum(
    "NonAsciiEnum",
    {"X\xe0": "x\xe0", "X\xe1": "x\xe1", "COMMA": "foo,bar"},
    type=str,
)


class IntTestEnum(int, Enum):
    ONE = 1
    TWO = 2
    THREE = 3


class CamelCaseEnum(int, Enum):
    one = 1
    TWO = 2
    Three = 3


enum_cases = [PureTestEnum, StrTestEnum, IntTestEnum]
encodings = ["utf8", crdb_encoding("latin1")]


@pytest.fixture(scope="session", autouse=True)
def make_test_enums(request, svcconn):
    for enum in enum_cases + [NonAsciiEnum]:
        ensure_enum(enum, svcconn)
    ensure_enum(CamelCaseEnum, svcconn, name="CamelCaseEnum")


def ensure_enum(enum, conn, name=""):
    if not name:
        name = enum.__name__.lower()
    labels = list(enum.__members__)
    conn.execute(sql.SQL("""
            drop type if exists {name};
            create type {name} as enum ({labels});
            """).format(name=sql.Identifier(name), labels=sql.SQL(",").join(labels)))
    return name, enum, labels


def test_fetch_info(conn):
    info = EnumInfo.fetch(conn, "StrTestEnum")
    assert info.name == "strtestenum"
    assert info.oid > 0
    assert info.oid != info.array_oid > 0
    assert len(info.labels) == len(StrTestEnum)
    assert info.labels == list(StrTestEnum.__members__)


async def test_fetch_info_async(aconn):
    info = await EnumInfo.fetch(aconn, "PureTestEnum")
    assert info.name == "puretestenum"
    assert info.oid > 0
    assert info.oid != info.array_oid > 0
    assert len(info.labels) == len(PureTestEnum)
    assert info.labels == list(PureTestEnum.__members__)


def test_register_makes_a_type(conn):
    info = EnumInfo.fetch(conn, "IntTestEnum")
    assert info
    assert info.enum is None
    register_enum(info, context=conn)
    assert info.enum is not None
    assert [e.name for e in info.enum] == list(IntTestEnum.__members__)


@pytest.mark.parametrize("enum", enum_cases)
@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_enum_loader(conn, enum, fmt_in, fmt_out):
    info = EnumInfo.fetch(conn, enum.__name__)
    register_enum(info, conn, enum=enum)

    for label in info.labels:
        cur = conn.execute(
            f"select %{fmt_in.value}::{enum.__name__}", [label], binary=fmt_out
        )
        assert cur.fetchone()[0] == enum[label]


@pytest.mark.parametrize("encoding", encodings)
@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_enum_loader_nonascii(conn, encoding, fmt_in, fmt_out):
    enum = NonAsciiEnum
    conn.execute(f"set client_encoding to {encoding}")

    info = EnumInfo.fetch(conn, enum.__name__)
    register_enum(info, conn, enum=enum)

    for label in info.labels:
        cur = conn.execute(
            f"select %{fmt_in.value}::{info.name}", [label], binary=fmt_out
        )
        assert cur.fetchone()[0] == enum[label]


@pytest.mark.crdb_skip("broken regtype")
@pytest.mark.parametrize("fmt_in", PyFormat)
def test_enum_quoted_name(conn, fmt_in):
    enum = CamelCaseEnum

    info = EnumInfo.fetch(conn, sql.Identifier(enum.__name__))
    register_enum(info, conn, enum=enum)

    cur = ClientCursor(conn)
    for value in enum:
        cur.execute(f"select %{fmt_in.value}", [value])
        assert next(cur)[0] is value


@pytest.mark.crdb_skip("encoding")
@pytest.mark.parametrize("enum", enum_cases)
@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_enum_loader_sqlascii(conn, enum, fmt_in, fmt_out):
    info = EnumInfo.fetch(conn, enum.__name__)
    register_enum(info, conn, enum)
    conn.execute("set client_encoding to sql_ascii")

    for label in info.labels:
        cur = conn.execute(
            f"select %{fmt_in.value}::{info.name}", [label], binary=fmt_out
        )
        assert cur.fetchone()[0] == enum[label]


@pytest.mark.parametrize("enum", enum_cases)
@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_enum_dumper(conn, enum, fmt_in, fmt_out):
    info = EnumInfo.fetch(conn, enum.__name__)
    register_enum(info, conn, enum)

    for item in enum:
        cur = conn.execute(f"select %{fmt_in.value}", [item], binary=fmt_out)
        assert cur.fetchone()[0] == item


@pytest.mark.parametrize("encoding", encodings)
@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_enum_dumper_nonascii(conn, encoding, fmt_in, fmt_out):
    enum = NonAsciiEnum
    conn.execute(f"set client_encoding to {encoding}")

    info = EnumInfo.fetch(conn, enum.__name__)
    register_enum(info, conn, enum)

    for item in enum:
        cur = conn.execute(f"select %{fmt_in.value}", [item], binary=fmt_out)
        assert cur.fetchone()[0] == item


@pytest.mark.crdb_skip("encoding")
@pytest.mark.parametrize("enum", enum_cases)
@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_enum_dumper_sqlascii(conn, enum, fmt_in, fmt_out):
    info = EnumInfo.fetch(conn, enum.__name__)
    register_enum(info, conn, enum)
    conn.execute("set client_encoding to sql_ascii")

    for item in enum:
        cur = conn.execute(f"select %{fmt_in.value}", [item], binary=fmt_out)
        assert cur.fetchone()[0] == item


@pytest.mark.parametrize("enum", enum_cases)
@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_generic_enum_dumper(conn, enum, fmt_in, fmt_out):
    for item in enum:
        if enum is PureTestEnum:
            want = item.name
        else:
            want = item.value

        cur = conn.execute(f"select %{fmt_in.value}", [item], binary=fmt_out)
        assert cur.fetchone()[0] == want


@pytest.mark.parametrize("encoding", encodings)
@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_generic_enum_dumper_nonascii(conn, encoding, fmt_in, fmt_out):
    conn.execute(f"set client_encoding to {encoding}")
    for item in NonAsciiEnum:
        cur = conn.execute(f"select %{fmt_in.value}", [item.value], binary=fmt_out)
        assert cur.fetchone()[0] == item.value


@pytest.mark.parametrize("enum", enum_cases)
@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_generic_enum_loader(conn, enum, fmt_in, fmt_out):
    for label in enum.__members__:
        cur = conn.execute(
            f"select %{fmt_in.value}::{enum.__name__}", [label], binary=fmt_out
        )
        want = enum[label].name
        if fmt_out == pq.Format.BINARY:
            want = want.encode()
        assert cur.fetchone()[0] == want


@pytest.mark.parametrize("encoding", encodings)
@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_generic_enum_loader_nonascii(conn, encoding, fmt_in, fmt_out):
    conn.execute(f"set client_encoding to {encoding}")

    for label in NonAsciiEnum.__members__:
        cur = conn.execute(
            f"select %{fmt_in.value}::nonasciienum", [label], binary=fmt_out
        )
        if fmt_out == pq.Format.TEXT:
            assert cur.fetchone()[0] == label
        else:
            assert cur.fetchone()[0] == label.encode(encoding)


@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_enum_array_loader(conn, fmt_in, fmt_out):
    enum = PureTestEnum
    info = EnumInfo.fetch(conn, enum.__name__)
    register_enum(info, conn, enum)

    labels = list(enum.__members__)
    cur = conn.execute(
        f"select %{fmt_in.value}::{info.name}[]", [labels], binary=fmt_out
    )
    assert cur.fetchone()[0] == list(enum)


@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_enum_array_dumper(conn, fmt_in, fmt_out):
    enum = StrTestEnum
    info = EnumInfo.fetch(conn, enum.__name__)
    register_enum(info, conn, enum)

    cur = conn.execute(f"select %{fmt_in.value}::text[]", [list(enum)], binary=fmt_out)
    assert cur.fetchone()[0] == list(enum.__members__)


@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("fmt_out", pq.Format)
def test_generic_enum_array_loader(conn, fmt_in, fmt_out):
    enum = IntTestEnum
    info = TypeInfo.fetch(conn, enum.__name__)
    info.register(conn)
    labels = list(enum.__members__)
    cur = conn.execute(
        f"select %{fmt_in.value}::{info.name}[]", [labels], binary=fmt_out
    )
    if fmt_out == pq.Format.TEXT:
        assert cur.fetchone()[0] == labels
    else:
        assert cur.fetchone()[0] == [item.encode() for item in labels]


def test_enum_error(conn):
    conn.autocommit = True

    info = EnumInfo.fetch(conn, "puretestenum")
    register_enum(info, conn, StrTestEnum)

    with pytest.raises(e.DataError):
        conn.execute("select %s::text", [StrTestEnum.ONE]).fetchone()
    with pytest.raises(e.DataError):
        conn.execute("select 'BAR'::puretestenum").fetchone()


@pytest.mark.parametrize("fmt_in", PyFormat)
@pytest.mark.parametrize("fmt_out", pq.Format)
@pytest.mark.parametrize(
    "mapping",
    [
        {StrTestEnum.ONE: "FOO", StrTestEnum.TWO: "BAR", StrTestEnum.THREE: "BAZ"},
        [
            (StrTestEnum.ONE, "FOO"),
            (StrTestEnum.TWO, "BAR"),
            (StrTestEnum.THREE, "BAZ"),
        ],
    ],
)
def test_remap(conn, fmt_in, fmt_out, mapping):
    info = EnumInfo.fetch(conn, "puretestenum")
    register_enum(info, conn, StrTestEnum, mapping=mapping)

    for member, label in [("ONE", "FOO"), ("TWO", "BAR"), ("THREE", "BAZ")]:
        cur = conn.execute(f"select %{fmt_in.value}::text", [StrTestEnum[member]])
        assert cur.fetchone()[0] == label
        cur = conn.execute(f"select '{label}'::puretestenum", binary=fmt_out)
        assert cur.fetchone()[0] is StrTestEnum[member]


def test_remap_rename(conn):
    RenamedEnum = Enum("RenamedEnum", "FOO BAR QUX")
    enum = RenamedEnum
    info = EnumInfo.fetch(conn, "puretestenum")
    register_enum(info, conn, enum, mapping={enum.QUX: "BAZ"})

    for member, label in [("FOO", "FOO"), ("BAR", "BAR"), ("QUX", "BAZ")]:
        cur = conn.execute("select %s::text", [enum[member]])
        assert cur.fetchone()[0] == label
        cur = conn.execute(f"select '{label}'::puretestenum")
        assert cur.fetchone()[0] is enum[member]


def test_remap_more_python(conn):
    LargerEnum = Enum("LargerEnum", "FOO BAR BAZ QUX QUUX QUUUX")
    enum = LargerEnum
    info = EnumInfo.fetch(conn, "puretestenum")
    mapping = {enum[m]: "BAZ" for m in ["QUX", "QUUX", "QUUUX"]}
    register_enum(info, conn, enum, mapping=mapping)

    for member, label in [("FOO", "FOO"), ("BAZ", "BAZ"), ("QUUUX", "BAZ")]:
        cur = conn.execute("select %s::text", [enum[member]])
        assert cur.fetchone()[0] == label

    for member, label in [("FOO", "FOO"), ("QUUUX", "BAZ")]:
        cur = conn.execute(f"select '{label}'::puretestenum")
        assert cur.fetchone()[0] is enum[member]


def test_remap_more_postgres(conn):
    SmallerEnum = Enum("SmallerEnum", "FOO")
    enum = SmallerEnum
    info = EnumInfo.fetch(conn, "puretestenum")
    mapping = [(enum.FOO, "BAR"), (enum.FOO, "BAZ")]
    register_enum(info, conn, enum, mapping=mapping)

    cur = conn.execute("select %s::text", [enum.FOO])
    assert cur.fetchone()[0] == "BAZ"

    for label in PureTestEnum.__members__:
        cur = conn.execute(f"select '{label}'::puretestenum")
        assert cur.fetchone()[0] is enum.FOO


def test_remap_by_value(conn):
    enum = Enum(  # type: ignore
        "ByValue",
        {m.lower(): m for m in PureTestEnum.__members__},
    )
    info = EnumInfo.fetch(conn, "puretestenum")
    register_enum(info, conn, enum, mapping={m: m.value for m in enum})

    for label in PureTestEnum.__members__:
        cur = conn.execute("select %s::text", [enum[label.lower()]])
        assert cur.fetchone()[0] == label

        cur = conn.execute(f"select '{label}'::puretestenum")
        assert cur.fetchone()[0] is enum[label.lower()]



from __future__ import annotations

import datetime as dt
import importlib
import ipaddress
from math import isnan
from uuid import UUID
from random import choice, random, randrange
from typing import Any
from decimal import Decimal
from contextlib import asynccontextmanager, contextmanager
from collections import deque

import pytest

import psycopg
from psycopg import sql
from psycopg.adapt import PyFormat
from psycopg.types.json import Json, Jsonb
from psycopg.types.range import Range
from psycopg.types.numeric import Int4, Int8
from psycopg.types.multirange import Multirange


@pytest.fixture
def faker(conn):
    return Faker(conn)


class Faker:
    """
    An object to generate random records.
    """

    json_max_level = 3
    json_max_length = 10
    str_max_length = 100
    list_max_length = 20
    tuple_max_length = 15

    def __init__(self, connection):
        self.conn = connection
        self.format = PyFormat.BINARY
        self.records = []

        self._schema = None
        self._types: list[type] | None = None
        self._types_names = None
        self._makers = {}
        self.table_name = sql.Identifier("fake_table")

    @property
    def schema(self):
        if not self._schema:
            self.schema = self.choose_schema()
        return self._schema

    @schema.setter
    def schema(self, schema):
        self._schema = schema
        self._types_names = None

    @property
    def fields_names(self):
        return [sql.Identifier(f"fld_{i}") for i in range(len(self.schema))]

    @property
    def types(self) -> list[type]:
        if not self._types:

            def key(cls: type) -> str:
                return f"{cls.__module__}.{cls.__qualname__}"

            self._types = sorted(self.get_supported_types(), key=key)

        return self._types

    @types.setter
    def types(self, types: list[type]) -> None:
        self._types = types

    @property
    def types_names_sql(self):
        if self._types_names:
            return self._types_names

        record = self.make_record(nulls=0)
        tx = psycopg.adapt.Transformer(self.conn)
        types = [
            self._get_type_name(tx, schema, value)
            for schema, value in zip(self.schema, record)
        ]
        self._types_names = types
        return types

    @property
    def types_names(self):
        types = [t.as_string(self.conn).replace('"', "") for t in self.types_names_sql]
        return types

    def _get_type_name(self, tx, schema, value):
        # Special case it as it is passed as unknown so is returned as text
        if schema == (list, str):
            return sql.SQL("text[]")

        registry = self.conn.adapters.types
        dumper = tx.get_dumper(value, self.format)
        dumper.dump(value)  # load the oid if it's dynamic (e.g. array)
        info = registry.get(dumper.oid) or registry.get("text")
        if dumper.oid == info.array_oid:
            return sql.SQL("{}[]").format(sql.Identifier(info.name))
        else:
            return sql.Identifier(info.name)

    @property
    def drop_stmt(self):
        return sql.SQL("drop table if exists {}").format(self.table_name)

    @property
    def create_stmt(self):
        field_values = []
        for name, type in zip(self.fields_names, self.types_names_sql):
            field_values.append(sql.SQL("{} {}").format(name, type))

        fields = sql.SQL(", ").join(field_values)
        return sql.SQL("create table {table} (id serial primary key, {fields})").format(
            table=self.table_name, fields=fields
        )

    @property
    def insert_stmt(self):
        phs = [sql.Placeholder(format=self.format) for i in range(len(self.schema))]
        return sql.SQL("insert into {} ({}) values ({})").format(
            self.table_name,
            sql.SQL(", ").join(self.fields_names),
            sql.SQL(", ").join(phs),
        )

    @property
    def select_stmt(self):
        fields = sql.SQL(", ").join(self.fields_names)
        return sql.SQL("select {} from {} order by id").format(fields, self.table_name)

    @contextmanager
    def find_insert_problem(self, conn):
        """Context manager to help finding a problematic value."""
        try:
            with conn.transaction():
                yield
        except psycopg.DatabaseError:
            cur = psycopg.Cursor(conn)
            # Repeat insert one field at time, until finding the wrong one
            cur.execute(self.drop_stmt)
            cur.execute(self.create_stmt)
            for i, rec in enumerate(self.records):
                for j, val in enumerate(rec):
                    try:
                        cur.execute(self._insert_field_stmt(j), (val,))
                    except psycopg.DatabaseError as e:
                        if len(r := repr(val)) > 200:
                            r = f"{r[:200]}... ({len(r)} chars)"
                        raise Exception(
                            f"value {r!r} at record {i} column0 {j} failed insert: {e}"
                        ) from None

            # just in case, but hopefully we should have triggered the problem
            raise

    @asynccontextmanager
    async def find_insert_problem_async(self, aconn):
        try:
            async with aconn.transaction():
                yield
        except psycopg.DatabaseError:
            acur = psycopg.AsyncCursor(aconn)
            # Repeat insert one field at time, until finding the wrong one
            await acur.execute(self.drop_stmt)
            await acur.execute(self.create_stmt)
            for i, rec in enumerate(self.records):
                for j, val in enumerate(rec):
                    try:
                        await acur.execute(self._insert_field_stmt(j), (val,))
                    except psycopg.DatabaseError as e:
                        if len(r := repr(val)) > 200:
                            r = f"{r[:200]}... ({len(r)} chars)"
                        raise Exception(
                            f"value {r!r} at record {i} column0 {j} failed insert: {e}"
                        ) from None

            # just in case, but hopefully we should have triggered the problem
            raise

    def _insert_field_stmt(self, i):
        ph = sql.Placeholder(format=self.format)
        return sql.SQL("insert into {} ({}) values ({})").format(
            self.table_name, self.fields_names[i], ph
        )

    def choose_schema(self, ncols=20):
        schema: list[tuple[type, ...] | type] = []
        while len(schema) < ncols:
            if (s := self.make_schema(choice(self.types))) is not None:
                schema.append(s)
        self.schema = schema
        return schema

    def make_records(self, nrecords):
        self.records = [self.make_record(nulls=0.05) for i in range(nrecords)]

    def make_record(self, nulls=0):
        if not nulls:
            return tuple(self.example(spec) for spec in self.schema)
        else:
            return tuple(
                self.make(spec) if random() > nulls else None for spec in self.schema
            )

    def assert_record(self, got, want):
        for spec, g, w in zip(self.schema, got, want):
            if g is None and w is None:
                continue
            m = self.get_matcher(spec)
            m(spec, g, w)

    def get_supported_types(self) -> set[type]:
        dumpers = self.conn.adapters._dumpers[self.format]
        rv = set()
        for cls in dumpers.keys():
            if isinstance(cls, str):
                if cls == "numpy.bool":
                    # An alias of numpy.bool_ for numpy > 2.
                    # Raises a warning in numpy > 1.20.
                    continue
                try:
                    cls = deep_import(cls)
                except ImportError:
                    continue
            if issubclass(cls, Multirange) and self.conn.info.server_version < 140000:
                continue

            rv.add(cls)

        # check all the types are handled
        for cls in rv:
            self.get_maker(cls)

        return rv

    def make_schema(self, cls: type) -> tuple[type, ...] | type | None:
        """Create a schema spec from a Python type.

        A schema specifies what Postgres type to generate when a Python type
        maps to more than one (e.g. tuple -> composite, list -> array[],
        datetime -> timestamp[tz]).

        A schema for a type is represented by a tuple (type, ...) which the
        matching make_*() method can interpret, or just type if the type
        doesn't require further specification.

        A `None` means that the type is not supported.
        """
        meth = self._get_method("schema", cls)
        return meth(cls) if meth else cls

    def get_maker(self, spec):
        cls = spec if isinstance(spec, type) else spec[0]

        try:
            return self._makers[cls]
        except KeyError:
            pass

        if meth := self._get_method("make", cls):
            self._makers[cls] = meth
            return meth
        else:
            raise NotImplementedError(f"cannot make fake objects of class {cls}")

    def get_matcher(self, spec):
        cls = spec if isinstance(spec, type) else spec[0]
        meth = self._get_method("match", cls)
        return meth if meth else self.match_any

    def _get_method(self, prefix, cls):
        name = cls.__name__
        if cls.__module__ != "builtins":
            name = f"{cls.__module__}.{name}"

        parts = name.split(".")
        for i in range(len(parts) - 1, -1, -1):
            if meth := getattr(self, f"{prefix}_{'_'.join(parts[-(i + 1):])}", None):
                return meth

        return None

    def make(self, spec):
        # spec can be a type or a tuple (type, options)
        return self.get_maker(spec)(spec)

    def example(self, spec):
        # A good representative of the object - no degenerate case
        cls = spec if isinstance(spec, type) else spec[0]
        if meth := self._get_method("example", cls):
            return meth(spec)
        else:
            return self.make(spec)

    def match_any(self, spec, got, want):
        assert got == want

    # methods to generate samples of specific types

    def make_Binary(self, spec):
        return self.make_bytes(spec)

    def match_Binary(self, spec, got, want):
        return want.obj == got

    def make_bool(self, spec):
        return spec(choice((True, False)))

    def make_bytearray(self, spec):
        return self.make_bytes(spec)

    def make_bytes(self, spec):
        length = randrange(self.str_max_length)
        return spec(bytes([randrange(256) for i in range(length)]))

    def make_date(self, spec):
        day = randrange(dt.date.max.toordinal())
        return dt.date.fromordinal(day + 1)

    def schema_datetime(self, cls):
        return self.schema_time(cls)

    def make_datetime(self, spec):
        # Add a day because with timezone we might go BC
        dtmin = dt.datetime.min + dt.timedelta(days=1)
        delta = dt.datetime.max - dtmin
        micros = randrange((delta.days + 1) * 24 * 60 * 60 * 1_000_000)
        rv = dtmin + dt.timedelta(microseconds=micros)
        if spec[1]:
            rv = rv.replace(tzinfo=self._make_tz(spec))
        return rv

    def match_datetime(self, spec, got, want):
        # Comparisons with different timezones is unreliable: certain pairs
        # are reported different but their delta is 0
        # https://bugs.python.org/issue45347
        assert not (got - want)

    def make_Decimal(self, spec):
        if random() >= 0.99:
            return Decimal(choice(self._decimal_special_values()))

        sign = choice("+-")
        num = choice(["0.zd", "d", "d.d"])
        while "z" in num:
            ndigits = randrange(1, 20)
            num = num.replace("z", "0" * ndigits, 1)
        while "d" in num:
            ndigits = randrange(1, 20)
            num = num.replace(
                "d", "".join([str(randrange(10)) for i in range(ndigits)]), 1
            )
        expsign = choice(["e+", "e-", ""])
        exp = randrange(20) if expsign else ""
        rv = Decimal(f"{sign}{num}{expsign}{exp}")
        return rv

    def match_Decimal(self, spec, got, want):
        if got is not None and got.is_nan():
            assert want.is_nan()
        else:
            assert got == want

    def _decimal_special_values(self):
        values = ["NaN", "sNaN"]

        if self.conn.info.vendor == "PostgreSQL":
            if self.conn.info.server_version >= 140000:
                values.extend(["Inf", "-Inf"])
        elif self.conn.info.vendor == "CockroachDB":
            if self.conn.info.server_version >= 220100:
                values.extend(["Inf", "-Inf"])
        else:
            pytest.fail(f"unexpected vendor: {self.conn.info.vendor}")

        return values

    def schema_Enum(self, cls):
        # TODO: can't fake those as we would need to create temporary types
        return None

    def make_Enum(self, spec):
        return None

    def make_float(self, spec, size=64):
        if random() <= 0.99:
            # These exponents should generate no inf/overflow
            if size == 64:
                s = f"{choice('-+')}0.{randrange(1 << 53)}e{randrange(-310, 309)}"
            elif size == 32:
                s = f"{choice('-+')}0.{randrange(1 << 22)}e{randrange(-37, 38)}"
            elif size == 16:
                s = f"{choice('-+')}0.{randrange(1 << 10)}e{randrange(-3, 4)}"
            else:
                assert False, size
            return spec(s)
        else:
            return choice(
                (spec(0.0), spec(-0.0), spec("-inf"), spec("inf"), spec("nan"))
            )

    def match_float(self, spec, got, want, rel=None):
        if got is not None and isnan(got):
            assert isnan(want)
        elif rel or self._server_rounds():
            assert got == pytest.approx(want, rel=rel)
        else:
            assert got == want

    def _server_rounds(self):
        """Return True if the connected server perform float rounding"""
        if self.conn.info.vendor == "CockroachDB":
            return True
        else:
            # Versions older than 12 make some rounding. e.g. in Postgres 10.4
            # select '-1.409006204063909e+112'::float8
            #      -> -1.40900620406391e+112
            return self.conn.info.server_version < 120000

    def make_Float4(self, spec):
        return self.make_float(spec, size=32)

    def match_Float4(self, spec, got, want):
        self.match_float(spec, got, want, rel=1e-5)

    def make_Float8(self, spec):
        return self.make_float(spec)

    match_Float8 = match_float

    def make_int(self, spec):
        return randrange(-(1 << 90), 1 << 90)

    def make_Int2(self, spec):
        return spec(randrange(-(1 << 15), 1 << 15))

    def make_Int4(self, spec):
        return spec(randrange(-(1 << 31), 1 << 31))

    def make_Int8(self, spec):
        return spec(randrange(-(1 << 63), 1 << 63))

    def make_IntNumeric(self, spec):
        return spec(randrange(-(1 << 100), 1 << 100))

    def make_IPv4Address(self, spec):
        return ipaddress.IPv4Address(bytes(randrange(256) for _ in range(4)))

    def make_IPv4Interface(self, spec):
        prefix = randrange(32)
        return ipaddress.IPv4Interface(
            (bytes(randrange(256) for _ in range(4)), prefix)
        )

    def make_IPv4Network(self, spec):
        return self.make_IPv4Interface(spec).network

    def make_IPv6Address(self, spec):
        return ipaddress.IPv6Address(bytes(randrange(256) for _ in range(16)))

    def make_IPv6Interface(self, spec):
        prefix = randrange(128)
        return ipaddress.IPv6Interface(
            (bytes(randrange(256) for _ in range(16)), prefix)
        )

    def make_IPv6Network(self, spec):
        return self.make_IPv6Interface(spec).network

    def make_Json(self, spec):
        return spec(self._make_json())

    def match_Json(self, spec, got, want):
        if want is not None:
            want = want.obj
        assert got == want

    def make_Jsonb(self, spec):
        return spec(self._make_json())

    def match_Jsonb(self, spec, got, want):
        self.match_Json(spec, got, want)

    def make_JsonFloat(self, spec):
        # A float limited to what json accepts
        # this exponent should generate no inf
        return float(f"{choice('-+')}0.{randrange(1 << 20)}e{randrange(-15, 15)}")

    def schema_list(self, cls):
        while True:
            # TODO: float lists are currently adapted as decimal.
            # There may be rounding errors or problems with inf.
            if (scls := choice(self.types)) is cls or scls is float:
                continue

            # CRDB doesn't support arrays of json
            # https://github.com/cockroachdb/cockroach/issues/23468
            if self.conn.info.vendor == "CockroachDB" and scls in (Json, Jsonb):
                continue
            if (schema := self.make_schema(scls)) is not None:
                break

        return (cls, schema)

    def make_list(self, spec):
        # don't make empty lists because they regularly fail cast
        length = randrange(1, self.list_max_length)
        spec = spec[1]
        while True:
            rv = [self.make(spec) for i in range(length)]

            # TODO multirange lists fail binary dump if the last element is
            # empty and there is no type annotation. See xfail in
            # test_multirange::test_dump_builtin_array
            if rv and isinstance(rv[-1], Multirange) and not rv[-1]:
                continue

            return rv

    def example_list(self, spec):
        return [self.example(spec[1])]

    def match_list(self, spec, got, want):
        assert len(got) == len(want)
        m = self.get_matcher(spec[1])
        for g, w in zip(got, want):
            m(spec[1], g, w)

    def make_memoryview(self, spec):
        return self.make_bytes(spec)

    def schema_Multirange(self, cls):
        return self.schema_Range(cls)

    def make_Multirange(self, spec, length=None, **kwargs):
        if length is None:
            length = randrange(0, self.list_max_length)

        def overlap(r1, r2):
            l1, u1 = r1.lower, r1.upper
            l2, u2 = r2.lower, r2.upper
            if l1 is None and l2 is None:
                return True
            elif l1 is None:
                l1 = l2
            elif l2 is None:
                l2 = l1

            if u1 is None and u2 is None:
                return True
            elif u1 is None:
                u1 = u2
            elif u2 is None:
                u2 = u1

            return l1 <= u2 and l2 <= u1

        out: list[Range[Any]] = []
        for i in range(length):
            if (r := self.make_Range((Range, spec[1]), **kwargs)).isempty:
                continue
            for r2 in out:
                if overlap(r, r2):
                    insert = False
                    break
            else:
                insert = True
            if insert:
                out.append(r)  # alternatively, we could merge

        return spec[0](sorted(out))

    def example_Multirange(self, spec):
        return self.make_Multirange(spec, length=1, empty_chance=0, no_bound_chance=0)

    def make_Int4Multirange(self, spec):
        return self.make_Multirange((spec, Int4))

    def make_Int8Multirange(self, spec):
        return self.make_Multirange((spec, Int8))

    def make_NumericMultirange(self, spec):
        return self.make_Multirange((spec, Decimal))

    def make_DateMultirange(self, spec):
        return self.make_Multirange((spec, dt.date))

    def make_TimestampMultirange(self, spec):
        return self.make_Multirange((spec, (dt.datetime, False)))

    def make_TimestamptzMultirange(self, spec):
        return self.make_Multirange((spec, (dt.datetime, True)))

    def match_Multirange(self, spec, got, want):
        assert len(got) == len(want)
        for ig, iw in zip(got, want):
            self.match_Range(spec, ig, iw)

    def match_Int4Multirange(self, spec, got, want):
        return self.match_Multirange((spec, Int4), got, want)

    def match_Int8Multirange(self, spec, got, want):
        return self.match_Multirange((spec, Int8), got, want)

    def match_NumericMultirange(self, spec, got, want):
        return self.match_Multirange((spec, Decimal), got, want)

    def match_DateMultirange(self, spec, got, want):
        return self.match_Multirange((spec, dt.date), got, want)

    def match_TimestampMultirange(self, spec, got, want):
        return self.match_Multirange((spec, (dt.datetime, False)), got, want)

    def match_TimestamptzMultirange(self, spec, got, want):
        return self.match_Multirange((spec, (dt.datetime, True)), got, want)

    def schema_NoneType(self, cls):
        return None

    def make_NoneType(self, spec):
        return None

    def make_Oid(self, spec):
        return spec(randrange(1 << 32))

    def schema_Range(self, cls):
        subtypes = [
            Decimal,
            Int4,
            Int8,
            dt.date,
            (dt.datetime, True),
            (dt.datetime, False),
        ]

        return (cls, choice(subtypes))

    def make_Range(self, spec, empty_chance=0.02, no_bound_chance=0.05):
        # TODO: drop format check after fixing binary dumping of empty ranges
        # (an array starting with an empty range will get the wrong type currently)
        if (
            random() < empty_chance
            and spec[0] is Range
            and self.format == PyFormat.TEXT
        ):
            return spec[0](empty=True)

        while True:
            bounds: list[Any] = []
            while len(bounds) < 2:
                if random() < no_bound_chance:
                    bounds.append(None)
                    continue

                val = self.make(spec[1])
                # NaN are allowed in a range, but comparison in Python get tricky.
                if spec[1] is Decimal and val.is_nan():
                    continue

                bounds.append(val)

            if bounds[0] is not None and bounds[1] is not None:
                if bounds[0] == bounds[1]:
                    # It would come out empty
                    continue

                if bounds[0] > bounds[1]:
                    bounds.reverse()

            # avoid generating ranges with no type info if dumping in binary
            # TODO: lift this limitation after test_copy_in_empty xfail is fixed
            if spec[0] is Range and self.format == PyFormat.BINARY:
                if bounds[0] is bounds[1] is None:
                    continue

            break

        r = spec[0](bounds[0], bounds[1], choice("[(") + choice("])"))
        return r

    def example_Range(self, spec):
        return self.make_Range(spec, empty_chance=0, no_bound_chance=0)

    def make_Int4Range(self, spec):
        return self.make_Range((spec, Int4))

    def make_Int8Range(self, spec):
        return self.make_Range((spec, Int8))

    def make_NumericRange(self, spec):
        return self.make_Range((spec, Decimal))

    def make_DateRange(self, spec):
        return self.make_Range((spec, dt.date))

    def make_TimestampRange(self, spec):
        return self.make_Range((spec, (dt.datetime, False)))

    def make_TimestamptzRange(self, spec):
        return self.make_Range((spec, (dt.datetime, True)))

    def match_Range(self, spec, got, want):
        # normalise the bounds of unbounded ranges
        if want.lower is None and want.lower_inc:
            want = type(want)(want.lower, want.upper, "(" + want.bounds[1])
        if want.upper is None and want.upper_inc:
            want = type(want)(want.lower, want.upper, want.bounds[0] + ")")

        # Normalise discrete ranges
        unit: dt.timedelta | int | None
        if spec[1] is dt.date:
            unit = dt.timedelta(days=1)
        elif type(spec[1]) is type and issubclass(spec[1], int):
            unit = 1
        else:
            unit = None

        if unit is not None:
            if want.lower is not None and not want.lower_inc:
                want = type(want)(want.lower + unit, want.upper, "[" + want.bounds[1])
            if want.upper_inc:
                want = type(want)(want.lower, want.upper + unit, want.bounds[0] + ")")

        if spec[1] == (dt.datetime, True) and not want.isempty:
            # work around https://bugs.python.org/issue45347
            def fix_dt(x):
                return x.astimezone(dt.timezone.utc) if x is not None else None

            def fix_range(r):
                return type(r)(fix_dt(r.lower), fix_dt(r.upper), r.bounds)

            want = fix_range(want)
            got = fix_range(got)

        assert got == want

    def match_Int4Range(self, spec, got, want):
        return self.match_Range((spec, Int4), got, want)

    def match_Int8Range(self, spec, got, want):
        return self.match_Range((spec, Int8), got, want)

    def match_NumericRange(self, spec, got, want):
        return self.match_Range((spec, Decimal), got, want)

    def match_DateRange(self, spec, got, want):
        return self.match_Range((spec, dt.date), got, want)

    def match_TimestampRange(self, spec, got, want):
        return self.match_Range((spec, (dt.datetime, False)), got, want)

    def match_TimestamptzRange(self, spec, got, want):
        return self.match_Range((spec, (dt.datetime, True)), got, want)

    def make_str(self, spec, length=0):
        if not length:
            length = randrange(self.str_max_length)

        rv: list[int] = []
        while len(rv) < length:
            c = randrange(1, 128) if random() < 0.5 else randrange(1, 0x110000)
            if not (0xD800 <= c <= 0xDBFF or 0xDC00 <= c <= 0xDFFF):
                rv.append(c)

        return "".join(map(chr, rv))

    def schema_time(self, cls):
        # Choose timezone yes/no
        return (cls, choice([True, False]))

    def make_time(self, spec):
        val = randrange(24 * 60 * 60 * 1_000_000)
        val, ms = divmod(val, 1_000_000)
        val, s = divmod(val, 60)
        h, m = divmod(val, 60)
        tz = self._make_tz(spec) if spec[1] else None
        return dt.time(h, m, s, ms, tz)

    CRDB_TIMEDELTA_MAX = dt.timedelta(days=1281239)

    def make_timedelta(self, spec):
        if self.conn.info.vendor == "CockroachDB":
            rng = [-self.CRDB_TIMEDELTA_MAX, self.CRDB_TIMEDELTA_MAX]
        else:
            rng = [dt.timedelta.min, dt.timedelta.max]

        return choice(rng) * random()

    def schema_tuple(self, cls):
        # TODO: this is a complicated matter as it would involve creating
        # temporary composite types.
        # length = randrange(1, self.tuple_max_length)
        # return (cls, self.make_random_schema(ncols=length))
        return None

    def make_tuple(self, spec):
        return tuple(self.make(s) for s in spec[1])

    def match_tuple(self, spec, got, want):
        assert len(got) == len(want) == len(spec[1])
        for g, w, s in zip(got, want, spec):
            if g is None or w is None:
                assert g is w
            else:
                m = self.get_matcher(s)
                m(s, g, w)

    def make_UUID(self, spec):
        return UUID(bytes=bytes([randrange(256) for i in range(16)]))

    def _make_json(self, container_chance=0.66):
        rec_types = [list, dict]
        scal_types = [type(None), int, JsonFloat, bool, str]
        if random() < container_chance:
            if (cls := choice(rec_types)) is list:
                return [
                    self._make_json(container_chance=container_chance / 2.0)
                    for i in range(randrange(self.json_max_length))
                ]
            elif cls is dict:
                return {
                    self.make_str(str, 15): self._make_json(
                        container_chance=container_chance / 2.0
                    )
                    for i in range(randrange(self.json_max_length))
                }
            else:
                assert False, f"unknown rec type: {cls}"

        else:
            cls = choice(scal_types)  # type: ignore[assignment]
            return self.make(cls)

    def _make_tz(self, spec):
        minutes = randrange(-12 * 60, 12 * 60 + 1)
        return dt.timezone(dt.timedelta(minutes=minutes))

    # numpy types support

    def make_numpy_bool_(self, spec):
        return self.make_bool(spec)

    def make_numpy_int8(self, spec):
        return spec(randrange(-(1 << 7), 1 << 7))

    def make_numpy_int16(self, spec):
        return self.make_Int2(spec)

    def make_numpy_int32(self, spec):
        return self.make_Int4(spec)

    def make_numpy_int64(self, spec):
        return self.make_Int8(spec)

    def make_numpy_longlong(self, spec):
        return self.make_numpy_int64(spec)

    def make_numpy_uint8(self, spec):
        return spec(randrange(0, 1 << 8))

    def make_numpy_uint16(self, spec):
        return spec(randrange(0, 1 << 16))

    def make_numpy_uint32(self, spec):
        return spec(randrange(0, 1 << 32))

    def make_numpy_uint64(self, spec):
        return spec(randrange(0, 1 << 64))

    def make_numpy_ulonglong(self, spec):
        return self.make_numpy_uint64(spec)

    def make_numpy_float16(self, spec):
        return self.make_float(spec, size=16)

    def make_numpy_float32(self, spec):
        return self.make_Float4(spec)

    def make_numpy_float64(self, spec):
        return self.make_Float8(spec)

    def match_numpy_ulonglong(self, spec, got, want):
        return self._match_numpy_with_decimal(spec, got, want)

    def match_numpy_uint64(self, spec, got, want):
        return self._match_numpy_with_decimal(spec, got, want)

    def _match_numpy_with_decimal(self, spec, got, want):
        assert isinstance(got, Decimal)
        return self.match_any(spec, int(got), want)

    def match_numpy_float16(self, spec, got, want):
        return self.match_float(spec, got, want, rel=1e-3)

    def match_numpy_float32(self, spec, got, want):
        return self.match_Float4(spec, got, want)

    def match_numpy_float64(self, spec, got, want):
        return self.match_Float8(spec, got, want)


class JsonFloat:
    pass


def deep_import(name):
    parts = deque(name.split("."))
    seen = []
    if not parts:
        raise ValueError("name must be a dot-separated name")

    seen.append(parts.popleft())
    thing = importlib.import_module(seen[-1])
    while parts:
        attr = parts.popleft()
        seen.append(attr)

        if hasattr(thing, attr):
            thing = getattr(thing, attr)
        else:
            thing = importlib.import_module(".".join(seen))

    return thing



from __future__ import annotations

import sys
import time
import signal
import asyncio
import threading
import subprocess as sp
from asyncio import create_task
from asyncio.queues import Queue

import pytest

import psycopg
from psycopg import errors as e


@pytest.mark.slow
async def test_commit_concurrency(aconn):
    # Check the condition reported in psycopg2#103
    # Because of bad status check, we commit even when a commit is already on
    # its way. We can detect this condition by the warnings.
    notices = Queue()  # type: ignore[var-annotated]
    aconn.add_notice_handler(lambda diag: notices.put_nowait(diag.message_primary))
    stop = False

    async def committer():
        while not stop:
            await aconn.commit()
            await asyncio.sleep(0)  # Allow the other worker to work

    async def runner():
        nonlocal stop
        cur = aconn.cursor()
        for i in range(1000):
            await cur.execute("select %s;", (i,))
            await aconn.commit()

        # Stop the committer thread
        stop = True

    await asyncio.gather(committer(), runner())
    assert notices.empty(), "%d notices raised" % notices.qsize()


@pytest.mark.slow
async def test_concurrent_execution(aconn_cls, dsn):
    async def worker():
        cnn = await aconn_cls.connect(dsn)
        cur = cnn.cursor()
        await cur.execute("select pg_sleep(0.5)")
        await cur.close()
        await cnn.close()

    workers = [worker(), worker()]
    t0 = time.time()
    await asyncio.gather(*workers)
    assert time.time() - t0 < 0.8, "something broken in concurrency"


async def canceller(aconn, errors):
    try:
        await asyncio.sleep(0.5)
        await aconn.cancel_safe()
    except Exception as exc:
        errors.append(exc)


@pytest.mark.slow
@pytest.mark.crdb_skip("cancel")
async def test_cancel(aconn):
    async def worker():
        cur = aconn.cursor()
        with pytest.raises(e.QueryCanceled):
            await cur.execute("select pg_sleep(2)")

    errors: list[Exception] = []
    workers = [worker(), canceller(aconn, errors)]

    t0 = time.time()
    await asyncio.gather(*workers)

    t1 = time.time()
    assert not errors
    assert 0.0 < t1 - t0 < 1.0

    # still working
    await aconn.rollback()
    cur = aconn.cursor()
    await cur.execute("select 1")
    assert await cur.fetchone() == (1,)


@pytest.mark.slow
@pytest.mark.crdb_skip("cancel")
async def test_cancel_stream(aconn):
    async def worker():
        cur = aconn.cursor()
        with pytest.raises(e.QueryCanceled):
            async for row in cur.stream("select pg_sleep(2)"):
                pass

    errors: list[Exception] = []
    workers = [worker(), canceller(aconn, errors)]

    t0 = time.time()
    await asyncio.gather(*workers)

    t1 = time.time()
    assert not errors
    assert 0.0 < t1 - t0 < 1.0

    # still working
    await aconn.rollback()
    cur = aconn.cursor()
    await cur.execute("select 1")
    assert await cur.fetchone() == (1,)


@pytest.mark.slow
@pytest.mark.timing
@pytest.mark.crdb_skip("pg_terminate_backend")
async def test_identify_closure(aconn_cls, dsn):
    async def closer():
        await asyncio.sleep(0.2)
        await conn2.execute(
            "select pg_terminate_backend(%s)", [aconn.pgconn.backend_pid]
        )

    aconn = await aconn_cls.connect(dsn)
    conn2 = await aconn_cls.connect(dsn)
    try:
        t = create_task(closer())
        t0 = time.time()
        try:
            with pytest.raises(psycopg.OperationalError):
                await aconn.execute("select pg_sleep(1.0)")
            t1 = time.time()
            assert 0.2 < t1 - t0 < 0.4
        finally:
            await asyncio.gather(t)
    finally:
        await aconn.close()
        await conn2.close()


@pytest.mark.slow
@pytest.mark.subprocess
@pytest.mark.skipif(
    sys.platform == "win32", reason="don't know how to Ctrl-C on Windows"
)
@pytest.mark.timing
@pytest.mark.crdb_skip("cancel")
def test_ctrl_c_handler(dsn):
    script = f"""\
import signal
import asyncio
import psycopg

async def main():
    ctrl_c = False
    loop = asyncio.get_event_loop()
    async with await psycopg.AsyncConnection.connect({dsn!r}) as conn:
        loop.add_signal_handler(signal.SIGINT, conn.cancel)
        cur = conn.cursor()
        try:
            await cur.execute("select pg_sleep(2)")
        except psycopg.errors.QueryCanceled:
            ctrl_c = True

        assert ctrl_c, "ctrl-c not received"
        assert (
            conn.info.transaction_status == psycopg.pq.TransactionStatus.INERROR
        ), f"transaction status: {{conn.info.transaction_status!r}}"

        await conn.rollback()
        assert (
            conn.info.transaction_status == psycopg.pq.TransactionStatus.IDLE
        ), f"transaction status: {{conn.info.transaction_status!r}}"

        await cur.execute("select 1")
        assert (await cur.fetchone()) == (1,)

asyncio.run(main())
"""
    if sys.platform == "win32":
        creationflags = sp.CREATE_NEW_PROCESS_GROUP
        sig = signal.CTRL_C_EVENT
    else:
        creationflags = 0
        sig = signal.SIGINT

    proc = sp.Popen([sys.executable, "-s", "-c", script], creationflags=creationflags)
    with pytest.raises(sp.TimeoutExpired):
        outs, errs = proc.communicate(timeout=1)

    proc.send_signal(sig)
    proc.communicate()
    assert proc.returncode == 0


@pytest.mark.slow
@pytest.mark.subprocess
@pytest.mark.skipif(
    sys.platform == "win32", reason="don't know how to Ctrl-C on Windows"
)
@pytest.mark.crdb("skip")
def test_ctrl_c(conn, dsn):
    # https://github.com/psycopg/psycopg/issues/543
    conn.autocommit = True

    APPNAME = "test_ctrl_c"
    script = f"""\
import asyncio
import psycopg

async def main():
    async with await psycopg.AsyncConnection.connect(
        {dsn!r}, application_name={APPNAME!r}
    ) as conn:
        await conn.execute("select pg_sleep(5)")

asyncio.run(main())
"""
    if sys.platform == "win32":
        creationflags = sp.CREATE_NEW_PROCESS_GROUP
        sig = signal.CTRL_C_EVENT
    else:
        creationflags = 0
        sig = signal.SIGINT

    proc = None

    def run_process():
        nonlocal proc
        proc = sp.Popen(
            [sys.executable, "-s", "-c", script],
            creationflags=creationflags,
            stderr=sp.PIPE,
        )
        proc.communicate()

    t = threading.Thread(target=run_process)
    t.start()

    for i in range(20):
        cur = conn.execute(
            "select pid from pg_stat_activity where application_name = %s", (APPNAME,)
        )

        if rec := cur.fetchone():
            pid = rec[0]
            break
        time.sleep(0.1)
    else:
        assert False, "process didn't start?"

    t0 = time.time()
    assert proc
    proc.send_signal(sig)
    proc.wait()

    for i in range(20):
        cur = conn.execute("select 1 from pg_stat_activity where pid = %s", (pid,))
        if not cur.fetchone():
            break
        time.sleep(0.1)
    else:
        assert False, "process didn't stop?"

    t1 = time.time()
    assert t1 - t0 < 1.0


@pytest.mark.slow
@pytest.mark.subprocess
@pytest.mark.parametrize("itimername, signame", [("ITIMER_REAL", "SIGALRM")])
def test_eintr(dsn, itimername, signame):
    try:
        itimer = int(getattr(signal, itimername))
        sig = int(getattr(signal, signame))
    except AttributeError:
        pytest.skip(f"unknown interrupt timer: {itimername}")

    script = f"""\
import signal
import asyncio
import psycopg

def signal_handler(signum, frame):
    assert signum == {sig!r}

# Install a handler for the signal
signal.signal({sig!r}, signal_handler)

# Restart system calls interrupted by the signal
signal.siginterrupt({sig!r}, False)


async def main():
    async with await psycopg.AsyncConnection.connect({dsn!r}) as conn:
        # Fire an interrupt signal every 0.25 seconds
        signal.setitimer({itimer!r}, 0.25, 0.25)

        cur = conn.cursor()
        await cur.execute("select 'ok' from pg_sleep(0.5)")
        print((await cur.fetchone())[0])

asyncio.run(main())
"""
    cp = sp.run(
        [sys.executable, "-s"], input=script, text=True, stdout=sp.PIPE, stderr=sp.PIPE
    )
    assert "InterruptedError" not in cp.stderr
    assert (
        cp.returncode == 0
    ), f"script terminated with {signal.Signals(abs(cp.returncode)).name}"
    assert cp.stdout.rstrip() == "ok"


@pytest.mark.slow
@pytest.mark.crdb("skip")
@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Fails with: An operation was attempted on something that is not a socket",
)
async def test_concurrent_close(dsn, aconn):
    # Test issue #608: concurrent closing shouldn't hang the server
    # (although, at the moment, it doesn't cancel a running query).
    pid = aconn.info.backend_pid
    await aconn.set_autocommit(True)

    async def worker():
        try:
            await aconn.execute("select pg_sleep(3)")
        except psycopg.OperationalError:
            pass  # expected

    t0 = time.time()
    task = create_task(worker())
    await asyncio.sleep(0.5)

    async def test():
        async with await psycopg.AsyncConnection.connect(dsn, autocommit=True) as conn1:
            cur = await conn1.execute(
                "select query from pg_stat_activity where pid = %s", [pid]
            )
            assert await cur.fetchone()
            await aconn.close()
            await asyncio.gather(task)
            await asyncio.sleep(0.5)
            t = time.time()
            # TODO: this statement can pass only if we send cancel on close
            # but because async cancelling is not available in the libpq,
            # we'd rather not do it.
            # cur = await conn1.execute(
            #     "select query from pg_stat_activity where pid = %s", [pid]
            # )
            # assert not await cur.fetchone()
            assert t - t0 < 2

    await asyncio.wait_for(test(), 5.0)


@pytest.mark.parametrize("what", ["commit", "rollback", "error"])
async def test_transaction_concurrency(aconn, what):
    await aconn.set_autocommit(True)

    evs = [asyncio.Event() for i in range(3)]

    async def worker(unlock, wait_on):
        with pytest.raises(e.ProgrammingError) as ex:
            async with aconn.transaction():
                unlock.set()
                await wait_on.wait()
                await aconn.execute("select 1")

                if what == "error":
                    1 / 0
                elif what == "rollback":
                    raise psycopg.Rollback()
                else:
                    assert what == "commit"

        if what == "error":
            assert "transaction rollback" in str(ex.value)
            assert isinstance(ex.value.__context__, ZeroDivisionError)
        elif what == "rollback":
            assert "transaction rollback" in str(ex.value)
            assert isinstance(ex.value.__context__, psycopg.Rollback)
        else:
            assert "transaction commit" in str(ex.value)

    # Start a first transaction in a task
    t1 = create_task(worker(unlock=evs[0], wait_on=evs[1]))
    await evs[0].wait()

    # Start a nested transaction in a task
    t2 = create_task(worker(unlock=evs[1], wait_on=evs[2]))

    # Terminate the first transaction before the second does
    await asyncio.gather(t1)
    evs[2].set()
    await asyncio.gather(t2)


@pytest.mark.slow
def test_type_error_shadow(dsn):
    script = f"""\
import sys
import uuid
import asyncio
import selectors

import psycopg

DSN = {dsn!r}
excs = []


async def connect() -> psycopg.AsyncConnection:
    return await psycopg.AsyncConnection.connect(DSN, autocommit=True)


async def worker(task_index: int, conn: psycopg.AsyncConnection) -> None:
    try:
        async with conn.cursor() as cur:
            await cur.executemany(
                "INSERT INTO lotsa_rows_table (id, task_index) VALUES (%s, %s)",
                [(uuid.uuid4(), task_index) for i in range(50000)],
            )
    except BaseException as be:
        excs.append(be)
        raise


async def crashing() -> None:
    1 / 0


async def main() -> None:
    async with await connect() as conn:
        await conn.execute("DROP TABLE IF EXISTS lotsa_rows_table")
        await conn.execute(
            "CREATE TABLE lotsa_rows_table (id uuid PRIMARY KEY, task_index INT)"
        )

    # create three tasks that each insert a large number of rows
    conns = [await connect() for i in range(3)]
    tasks = [asyncio.create_task(worker(i, conn=conns[i])) for i in range(3)]

    # create one more task that fails immediately
    tasks.append(asyncio.create_task(crashing()))

    try:
        await asyncio.gather(*tasks)
    except ZeroDivisionError:
        pass
    else:
        raise AssertionError("ValueError not raised")


if __name__ == "__main__":
    kwargs = {{}}
    if sys.platform == "win32":
        if sys.version_info >= (3, 12):
            kwargs["loop_factory"] = lambda: asyncio.SelectorEventLoop(
                selectors.SelectSelector()
            )
        else:
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    try:
        asyncio.run(main(), **kwargs)
    finally:
        assert excs, "No exception raised by workers"
        for ex in excs:
            assert isinstance(ex, asyncio.CancelledError), f"raised {{ex!r}} instead"
"""

    sp.run([sys.executable, "-s"], input=script, text=True, check=True)



# type: ignore
"""
ticket role
~~~~~~~~~~~

An interpreted text role to link docs to tickets issues.

:copyright: Copyright 2013 by Daniele Varrazzo.
:copyright: Copyright 2021 The Psycopg Team
"""

import re

from docutils import nodes, utils
from docutils.parsers.rst import roles


def ticket_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    if (cfg := inliner.document.settings.env.app.config).ticket_url is None:
        msg = inliner.reporter.warning(
            "ticket not configured: please configure ticket_url in conf.py"
        )
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]

    rv = [nodes.Text(name + " ")]
    tokens = re.findall(r"(#?\d+)|([^\d#]+)", text)
    for ticket, noise in tokens:
        if ticket:
            num = int(ticket.replace("#", ""))

            url = cfg.ticket_url % num
            roles.set_classes(options)
            node = nodes.reference(
                ticket, utils.unescape(ticket), refuri=url, **options
            )

            rv.append(node)

        else:
            assert noise
            rv.append(nodes.Text(noise))

    return rv, []


def setup(app):
    app.add_config_value("ticket_url", None, "env")
    app.add_role("ticket", ticket_role)
    app.add_role("tickets", ticket_role)



"""
Utilities to ease the differences between async and sync code.

These object offer a similar interface between sync and async versions; the
script async_to_sync.py will replace the async names with the sync names
when generating the sync version.
"""

# Copyright (C) 2023 The Psycopg Team

from __future__ import annotations

import time
import queue
import asyncio
import logging
import threading
from typing import Any, Awaitable, ParamSpec, TypeAlias, overload
from inspect import isawaitable
from collections.abc import Callable, Coroutine

from ._compat import TypeVar

logger = logging.getLogger("psycopg.pool")
T = TypeVar("T")
P = ParamSpec("P")

# Re-exports
Event = threading.Event
Condition = threading.Condition
Lock = threading.RLock
ALock = asyncio.Lock
sleep = time.sleep

Worker: TypeAlias = threading.Thread
AWorker: TypeAlias = asyncio.Task[None]


def current_thread_name() -> str:
    return threading.current_thread().name


def current_task_name() -> str:
    t = asyncio.current_task()
    return t.get_name() if t else "<no task>"


class Queue(queue.Queue[T]):
    """
    A Queue subclass with an interruptible get() method.
    """

    def get(self, block: bool = True, timeout: float | None = None) -> T:
        # Always specify a timeout to make the wait interruptible.
        if timeout is None:
            timeout = 24.0 * 60.0 * 60.0
        return super().get(block, timeout)


class AEvent(asyncio.Event):
    """
    Subclass of asyncio.Event adding a wait with timeout like threading.Event.

    wait_timeout() is converted to wait() by async_to_sync.
    """

    async def wait_timeout(self, timeout: float) -> bool:
        try:
            await asyncio.wait_for(self.wait(), timeout)
            return True
        except asyncio.TimeoutError:
            return False


class ACondition(asyncio.Condition):
    """
    Subclass of asyncio.Condition adding a wait with timeout like threading.Condition.

    wait_timeout() is converted to wait() by async_to_sync.
    """

    async def wait_timeout(self, timeout: float) -> bool:
        try:
            await asyncio.wait_for(self.wait(), timeout)
            return True
        except asyncio.TimeoutError:
            return False


class AQueue(asyncio.Queue[T]):
    pass


def aspawn(
    f: Callable[..., Coroutine[Any, Any, None]],
    args: tuple[Any, ...] = (),
    name: str | None = None,
) -> asyncio.Task[None]:
    """
    Equivalent to asyncio.create_task.
    """
    return asyncio.create_task(f(*args), name=name)


def spawn(
    f: Callable[..., Any],
    args: tuple[Any, ...] = (),
    name: str | None = None,
) -> threading.Thread:
    """
    Equivalent to creating and running a daemon thread.
    """
    t = threading.Thread(target=f, args=args, name=name, daemon=True)
    t.start()
    return t


async def agather(
    *tasks: asyncio.Task[Any], timeout: float | None = None, timeout_hint: str = ""
) -> None:
    """
    Equivalent to asyncio.gather or Thread.join()
    """
    wait = asyncio.gather(*tasks)
    try:
        if timeout is not None:
            await asyncio.wait_for(asyncio.shield(wait), timeout=timeout)
        else:
            await wait
    except asyncio.TimeoutError:
        pass
    else:
        return

    for t in tasks:
        if t.done():
            continue
        logger.warning("couldn't stop task %r within %s seconds", t.get_name(), timeout)
        if timeout_hint:
            logger.warning("hint: %s", timeout_hint)


def gather(
    *tasks: threading.Thread, timeout: float | None = None, timeout_hint: str = ""
) -> None:
    """
    Equivalent to asyncio.gather or Thread.join()
    """
    for t in tasks:
        if not t.is_alive():
            continue
        t.join(timeout)
        if not t.is_alive():
            continue
        logger.warning("couldn't stop thread %r within %s seconds", t.name, timeout)
        if timeout_hint:
            logger.warning("hint: %s", timeout_hint)


def asleep(seconds: float) -> Coroutine[Any, Any, None]:
    """
    Equivalent to asyncio.sleep(), converted to time.sleep() by async_to_sync.
    """
    return asyncio.sleep(seconds)


@overload
async def ensure_async(
    f: Callable[P, Awaitable[T]], *args: P.args, **kwargs: P.kwargs
) -> T: ...


@overload
async def ensure_async(f: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T: ...


async def ensure_async(
    f: Callable[P, T] | Callable[P, Coroutine[Any, Any, T]],
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    rv = f(*args, **kwargs)
    if isawaitable(rv):
        rv = await rv
    return rv



"""
psycopg exceptions

DBAPI-defined Exceptions are defined in the following hierarchy::

    Exceptions
    |__Warning
    |__Error
       |__InterfaceError
       |__DatabaseError
          |__DataError
          |__OperationalError
          |__IntegrityError
          |__InternalError
          |__ProgrammingError
          |__NotSupportedError
"""

# Copyright (C) 2020 The Psycopg Team

from __future__ import annotations

from typing import TYPE_CHECKING, Any, NoReturn, TypeAlias, TypeGuard
from asyncio import CancelledError
from dataclasses import dataclass, field, fields
from collections.abc import Callable, Sequence

from .pq.abc import PGconn, PGresult
from .pq._enums import ConnStatus, DiagnosticField, PipelineStatus, TransactionStatus

if TYPE_CHECKING:
    from .pq.misc import ConninfoOption, PGnotify

ErrorInfo: TypeAlias = PGresult | dict[int, bytes | None] | None

_sqlcodes: dict[str, type[Error]] = {}


@dataclass
class FinishedPGconn:
    """Finished libpq connection.

    Attributes are set from a real `~psycopg.pq.PGconn` but any operations will
    raise an `~psycopg.OperationalError`.
    """

    info: list[ConninfoOption] = field(default_factory=list)

    db: bytes = b""
    user: bytes = b""
    password: bytes = b""
    host: bytes = b""
    hostaddr: bytes = b""
    port: bytes = b""
    tty: bytes = b""
    options: bytes = b""
    status: int = ConnStatus.BAD.value
    transaction_status: int = TransactionStatus.UNKNOWN.value
    pipeline_status: int = PipelineStatus.OFF.value

    error_message: bytes = b""
    _encoding: str = "utf-8"
    protocol_version: int = 0
    full_protocol_version: int = 0
    server_version: int = 0

    backend_pid: int = 0
    needs_password: bool = False
    used_password: bool = False
    used_gssapi: bool = False
    ssl_in_use: bool = False

    nonblocking: int = 0

    notice_handler: Callable[[PGresult], None] | None = None
    notify_handler: Callable[[PGnotify], None] | None = None

    @staticmethod
    def _raise() -> NoReturn:
        raise OperationalError("the connection is closed")

    @classmethod
    def connect(cls, *args: Any) -> NoReturn:
        raise TypeError(f"{cls} is unusable")

    @classmethod
    def connect_start(cls, *args: Any) -> NoReturn:
        raise TypeError(f"{cls} is unusable")

    def connect_poll(self) -> NoReturn:
        self._raise()

    def finish(self) -> None:
        pass

    def reset(self) -> NoReturn:
        self._raise()

    def get_error_message(self, encoding: str = "") -> str:
        return "the connection is closed"

    def reset_start(self) -> NoReturn:
        self._raise()

    def reset_poll(self) -> NoReturn:
        self._raise()

    @classmethod
    def ping(cls, *args: Any) -> NoReturn:
        raise TypeError(f"{cls} is unusable")

    def parameter_status(self, *args: Any) -> NoReturn:
        self._raise()

    @property
    def socket(self) -> NoReturn:
        self._raise()

    def exec_(self, *args: Any) -> NoReturn:
        self._raise()

    def send_query(self, *args: Any) -> None:
        self._raise()

    def exec_params(self, *args: Any) -> NoReturn:
        self._raise()

    def send_query_params(self, *args: Any) -> NoReturn:
        self._raise()

    def send_prepare(self, *args: Any) -> NoReturn:
        self._raise()

    def send_query_prepared(self, *args: Any) -> NoReturn:
        self._raise()

    def prepare(self, *args: Any) -> NoReturn:
        self._raise()

    def exec_prepared(self, *args: Any) -> NoReturn:
        self._raise()

    def describe_prepared(self, *args: Any) -> NoReturn:
        self._raise()

    def send_describe_prepared(self, *args: Any) -> NoReturn:
        self._raise()

    def describe_portal(self, *args: Any) -> NoReturn:
        self._raise()

    def send_describe_portal(self, *args: Any) -> NoReturn:
        self._raise()

    def close_prepared(self, *args: Any) -> NoReturn:
        self._raise()

    def send_close_prepared(self, *args: Any) -> NoReturn:
        self._raise()

    def close_portal(self, *args: Any) -> NoReturn:
        self._raise()

    def send_close_portal(self, *args: Any) -> NoReturn:
        self._raise()

    def get_result(self) -> NoReturn:
        self._raise()

    def consume_input(self) -> NoReturn:
        self._raise()

    def is_busy(self) -> NoReturn:
        self._raise()

    def flush(self) -> NoReturn:
        self._raise()

    def set_single_row_mode(self) -> NoReturn:
        self._raise()

    def set_chunked_rows_mode(self, size: int) -> NoReturn:
        self._raise()

    def cancel_conn(self) -> NoReturn:
        self._raise()

    def get_cancel(self) -> NoReturn:
        self._raise()

    def notifies(self) -> NoReturn:
        self._raise()

    def put_copy_data(self, *args: Any) -> NoReturn:
        self._raise()

    def put_copy_end(self, *args: Any) -> NoReturn:
        self._raise()

    def get_copy_data(self, *args: Any) -> NoReturn:
        self._raise()

    def trace(self, *args: Any) -> NoReturn:
        self._raise()

    def set_trace_flags(self, *args: Any) -> NoReturn:
        self._raise()

    def untrace(self) -> NoReturn:
        self._raise()

    def encrypt_password(self, *args: Any) -> NoReturn:
        self._raise()

    def change_password(self, *args: Any) -> NoReturn:
        self._raise()

    def make_empty_result(self, *args: Any) -> NoReturn:
        self._raise()

    def enter_pipeline_mode(self) -> NoReturn:
        self._raise()

    def exit_pipeline_mode(self) -> NoReturn:
        self._raise()

    def pipeline_sync(self) -> NoReturn:
        self._raise()

    def send_flush_request(self) -> NoReturn:
        self._raise()


def finish_pgconn(pgconn: PGconn) -> PGconn:
    args = {}
    for f in fields(FinishedPGconn):
        try:
            args[f.name] = getattr(pgconn, f.name)
        except Exception:
            pass
    pgconn.finish()
    return FinishedPGconn(**args)


class Warning(Exception):
    """
    Exception raised for important warnings.

    Defined for DBAPI compatibility, but never raised by ``psycopg``.
    """

    __module__ = "psycopg"


class Error(Exception):
    """
    Base exception for all the errors psycopg will raise.

    Exception that is the base class of all other error exceptions. You can
    use this to catch all errors with one single `!except` statement.

    This exception is guaranteed to be picklable.
    """

    __module__ = "psycopg"

    sqlstate: str | None = None

    def __init__(
        self,
        *args: Sequence[Any],
        info: ErrorInfo = None,
        encoding: str = "utf-8",
        pgconn: PGconn | None = None,
    ):
        super().__init__(*args)
        self._info = info
        self._encoding = encoding
        self._pgconn = pgconn

        # Handle sqlstate codes for which we don't have a class.
        if not self.sqlstate and info:
            self.sqlstate = self.diag.sqlstate

    @property
    def pgconn(self) -> PGconn | None:
        """The connection object, if the error was raised from a connection attempt.

        :rtype: psycopg.pq.PGconn | None
        """
        return self._pgconn if self._pgconn else None

    @property
    def pgresult(self) -> PGresult | None:
        """The result object, if the exception was raised after a failed query.

        :rtype: psycopg.pq.PGresult | None
        """
        return self._info if _is_pgresult(self._info) else None

    @property
    def diag(self) -> Diagnostic:
        """
        A `Diagnostic` object to inspect details of the errors from the database.
        """
        return Diagnostic(self._info, encoding=self._encoding)

    def __reduce__(self) -> str | tuple[Any, ...]:
        res = super().__reduce__()
        if isinstance(res, tuple) and len(res) >= 3:
            # To make the exception picklable
            res[2]["_info"] = _info_to_dict(self._info)
            res[2]["_pgconn"] = None

        return res


class InterfaceError(Error):
    """
    An error related to the database interface rather than the database itself.
    """

    __module__ = "psycopg"


class DatabaseError(Error):
    """
    Exception raised for errors that are related to the database.
    """

    __module__ = "psycopg"

    def __init_subclass__(cls, code: str | None = None, name: str | None = None):
        if code:
            _sqlcodes[code] = cls
            cls.sqlstate = code
        if name:
            _sqlcodes[name] = cls


class DataError(DatabaseError):
    """
    An error caused by problems with the processed data.

    Examples may be division by zero, numeric value out of range, etc.
    """

    __module__ = "psycopg"


class OperationalError(DatabaseError):
    """
    An error related to the database's operation.

    These errors are not necessarily under the control of the programmer, e.g.
    an unexpected disconnect occurs, the data source name is not found, a
    transaction could not be processed, a memory allocation error occurred
    during processing, etc.
    """

    __module__ = "psycopg"


class IntegrityError(DatabaseError):
    """
    An error caused when the relational integrity of the database is affected.

    An example may be a foreign key check failed.
    """

    __module__ = "psycopg"


class InternalError(DatabaseError):
    """
    An error generated when the database encounters an internal error,

    Examples could be the cursor is not valid anymore, the transaction is out
    of sync, etc.
    """

    __module__ = "psycopg"


class ProgrammingError(DatabaseError):
    """
    Exception raised for programming errors

    Examples may be table not found or already exists, syntax error in the SQL
    statement, wrong number of parameters specified, etc.
    """

    __module__ = "psycopg"


class NotSupportedError(DatabaseError):
    """
    A method or database API was used which is not supported by the database.
    """

    __module__ = "psycopg"


class ConnectionTimeout(OperationalError):
    """
    Exception raised on timeout of the `~psycopg.Connection.connect()` method.

    The error is raised if the ``connect_timeout`` is specified and a
    connection is not obtained in useful time.

    Subclass of `~psycopg.OperationalError`.
    """


class CancellationTimeout(OperationalError):
    """
    Exception raised on timeout of connection's
    `~psycopg.Connection.cancel_safe()` method.

    Subclass of `~psycopg.OperationalError`.
    """


class PipelineAborted(OperationalError):
    """
    Raised when a operation fails because the current pipeline is in aborted state.

    Subclass of `~psycopg.OperationalError`.
    """


class Diagnostic:
    """Details from a database error report."""

    def __init__(self, info: ErrorInfo, encoding: str = "utf-8"):
        self._info = info
        self._encoding = encoding

    @property
    def severity(self) -> str | None:
        return self._error_message(DiagnosticField.SEVERITY)

    @property
    def severity_nonlocalized(self) -> str | None:
        return self._error_message(DiagnosticField.SEVERITY_NONLOCALIZED)

    @property
    def sqlstate(self) -> str | None:
        return self._error_message(DiagnosticField.SQLSTATE)

    @property
    def message_primary(self) -> str | None:
        return self._error_message(DiagnosticField.MESSAGE_PRIMARY)

    @property
    def message_detail(self) -> str | None:
        return self._error_message(DiagnosticField.MESSAGE_DETAIL)

    @property
    def message_hint(self) -> str | None:
        return self._error_message(DiagnosticField.MESSAGE_HINT)

    @property
    def statement_position(self) -> str | None:
        return self._error_message(DiagnosticField.STATEMENT_POSITION)

    @property
    def internal_position(self) -> str | None:
        return self._error_message(DiagnosticField.INTERNAL_POSITION)

    @property
    def internal_query(self) -> str | None:
        return self._error_message(DiagnosticField.INTERNAL_QUERY)

    @property
    def context(self) -> str | None:
        return self._error_message(DiagnosticField.CONTEXT)

    @property
    def schema_name(self) -> str | None:
        return self._error_message(DiagnosticField.SCHEMA_NAME)

    @property
    def table_name(self) -> str | None:
        return self._error_message(DiagnosticField.TABLE_NAME)

    @property
    def column_name(self) -> str | None:
        return self._error_message(DiagnosticField.COLUMN_NAME)

    @property
    def datatype_name(self) -> str | None:
        return self._error_message(DiagnosticField.DATATYPE_NAME)

    @property
    def constraint_name(self) -> str | None:
        return self._error_message(DiagnosticField.CONSTRAINT_NAME)

    @property
    def source_file(self) -> str | None:
        return self._error_message(DiagnosticField.SOURCE_FILE)

    @property
    def source_line(self) -> str | None:
        return self._error_message(DiagnosticField.SOURCE_LINE)

    @property
    def source_function(self) -> str | None:
        return self._error_message(DiagnosticField.SOURCE_FUNCTION)

    def _error_message(self, field: DiagnosticField) -> str | None:
        if self._info:
            if isinstance(self._info, dict):
                val = self._info.get(field)
            else:
                val = self._info.error_field(field)

            if val is not None:
                return val.decode(self._encoding, "replace")

        return None

    def __reduce__(self) -> str | tuple[Any, ...]:
        res = super().__reduce__()
        if isinstance(res, tuple) and len(res) >= 3:
            res[2]["_info"] = _info_to_dict(self._info)

        return res


def _info_to_dict(info: ErrorInfo) -> ErrorInfo:
    """
    Convert a PGresult to a dictionary to make the info picklable.
    """
    # PGresult is a protocol, can't use isinstance
    if _is_pgresult(info):
        return {v: info.error_field(v) for v in DiagnosticField}
    else:
        return info


def lookup(sqlstate: str) -> type[Error]:
    """Lookup an error code or `constant name`__ and return its exception class.

    Raise `!KeyError` if the code is not found.

    .. __: https://www.postgresql.org/docs/current/errcodes-appendix.html
            #ERRCODES-TABLE
    """
    return _sqlcodes[sqlstate.upper()]


def error_from_result(result: PGresult, encoding: str = "utf-8") -> Error:
    state = result.error_field(DiagnosticField.SQLSTATE) or b""
    cls = _class_for_state(state.decode("utf-8", "replace"))
    return cls(result.get_error_message(encoding), info=result, encoding=encoding)


def _is_pgresult(info: ErrorInfo) -> TypeGuard[PGresult]:
    """Return True if an ErrorInfo is a PGresult instance."""
    # PGresult is a protocol, can't use isinstance
    return hasattr(info, "error_field")


def _class_for_state(sqlstate: str) -> type[Error]:
    try:
        return lookup(sqlstate)
    except KeyError:
        return get_base_exception(sqlstate)


def get_base_exception(sqlstate: str) -> type[Error]:
    return (
        _base_exc_map.get(sqlstate[:2])
        or _base_exc_map.get(sqlstate[:1])
        or DatabaseError
    )


_base_exc_map = {
    "08": OperationalError,  # Connection Exception
    "0A": NotSupportedError,  # Feature Not Supported
    "10": ProgrammingError,  # XQuery Error
    "20": ProgrammingError,  # Case Not Foud
    "21": ProgrammingError,  # Cardinality Violation
    "22": DataError,  # Data Exception
    "23": IntegrityError,  # Integrity Constraint Violation
    "24": InternalError,  # Invalid Cursor State
    "25": InternalError,  # Invalid Transaction State
    "26": ProgrammingError,  # Invalid SQL Statement Name *
    "27": OperationalError,  # Triggered Data Change Violation
    "28": OperationalError,  # Invalid Authorization Specification
    "2B": InternalError,  # Dependent Privilege Descriptors Still Exist
    "2D": InternalError,  # Invalid Transaction Termination
    "2F": OperationalError,  # SQL Routine Exception *
    "34": ProgrammingError,  # Invalid Cursor Name *
    "38": OperationalError,  # External Routine Exception *
    "39": OperationalError,  # External Routine Invocation Exception *
    "3B": OperationalError,  # Savepoint Exception *
    "3D": ProgrammingError,  # Invalid Catalog Name
    "3F": ProgrammingError,  # Invalid Schema Name
    "40": OperationalError,  # Transaction Rollback
    "42": ProgrammingError,  # Syntax Error or Access Rule Violation
    "44": ProgrammingError,  # WITH CHECK OPTION Violation
    "53": OperationalError,  # Insufficient Resources
    "54": OperationalError,  # Program Limit Exceeded
    "55": OperationalError,  # Object Not In Prerequisite State
    "57": OperationalError,  # Operator Intervention
    "58": OperationalError,  # System Error (errors external to PostgreSQL itself)
    "F": OperationalError,  # Configuration File Error
    "H": OperationalError,  # Foreign Data Wrapper Error (SQL/MED)
    "P": ProgrammingError,  # PL/pgSQL Error
    "X": InternalError,  # Internal Error
}


# Error classes generated by tools/update_errors.py

# fmt: off
# autogenerated: start


# Class 02 - No Data (this is also a warning class per the SQL standard)

class NoData(DatabaseError,
    code='02000', name='NO_DATA'):
    pass

class NoAdditionalDynamicResultSetsReturned(DatabaseError,
    code='02001', name='NO_ADDITIONAL_DYNAMIC_RESULT_SETS_RETURNED'):
    pass


# Class 03 - SQL Statement Not Yet Complete

class SqlStatementNotYetComplete(DatabaseError,
    code='03000', name='SQL_STATEMENT_NOT_YET_COMPLETE'):
    pass


# Class 08 - Connection Exception

class ConnectionException(OperationalError,
    code='08000', name='CONNECTION_EXCEPTION'):
    pass

class SqlclientUnableToEstablishSqlconnection(OperationalError,
    code='08001', name='SQLCLIENT_UNABLE_TO_ESTABLISH_SQLCONNECTION'):
    pass

class ConnectionDoesNotExist(OperationalError,
    code='08003', name='CONNECTION_DOES_NOT_EXIST'):
    pass

class SqlserverRejectedEstablishmentOfSqlconnection(OperationalError,
    code='08004', name='SQLSERVER_REJECTED_ESTABLISHMENT_OF_SQLCONNECTION'):
    pass

class ConnectionFailure(OperationalError,
    code='08006', name='CONNECTION_FAILURE'):
    pass

class TransactionResolutionUnknown(OperationalError,
    code='08007', name='TRANSACTION_RESOLUTION_UNKNOWN'):
    pass

class ProtocolViolation(OperationalError,
    code='08P01', name='PROTOCOL_VIOLATION'):
    pass


# Class 09 - Triggered Action Exception

class TriggeredActionException(DatabaseError,
    code='09000', name='TRIGGERED_ACTION_EXCEPTION'):
    pass


# Class 0A - Feature Not Supported

class FeatureNotSupported(NotSupportedError,
    code='0A000', name='FEATURE_NOT_SUPPORTED'):
    pass


# Class 0B - Invalid Transaction Initiation

class InvalidTransactionInitiation(DatabaseError,
    code='0B000', name='INVALID_TRANSACTION_INITIATION'):
    pass


# Class 0F - Locator Exception

class LocatorException(DatabaseError,
    code='0F000', name='LOCATOR_EXCEPTION'):
    pass

class InvalidLocatorSpecification(DatabaseError,
    code='0F001', name='INVALID_LOCATOR_SPECIFICATION'):
    pass


# Class 0L - Invalid Grantor

class InvalidGrantor(DatabaseError,
    code='0L000', name='INVALID_GRANTOR'):
    pass

class InvalidGrantOperation(DatabaseError,
    code='0LP01', name='INVALID_GRANT_OPERATION'):
    pass


# Class 0P - Invalid Role Specification

class InvalidRoleSpecification(DatabaseError,
    code='0P000', name='INVALID_ROLE_SPECIFICATION'):
    pass


# Class 0Z - Diagnostics Exception

class DiagnosticsException(DatabaseError,
    code='0Z000', name='DIAGNOSTICS_EXCEPTION'):
    pass

class StackedDiagnosticsAccessedWithoutActiveHandler(DatabaseError,
    code='0Z002', name='STACKED_DIAGNOSTICS_ACCESSED_WITHOUT_ACTIVE_HANDLER'):
    pass


# Class 10 - XQuery Error

class InvalidArgumentForXquery(ProgrammingError,
    code='10608', name='INVALID_ARGUMENT_FOR_XQUERY'):
    pass


# Class 20 - Case Not Found

class CaseNotFound(ProgrammingError,
    code='20000', name='CASE_NOT_FOUND'):
    pass


# Class 21 - Cardinality Violation

class CardinalityViolation(ProgrammingError,
    code='21000', name='CARDINALITY_VIOLATION'):
    pass


# Class 22 - Data Exception

class DataException(DataError,
    code='22000', name='DATA_EXCEPTION'):
    pass

class StringDataRightTruncation(DataError,
    code='22001', name='STRING_DATA_RIGHT_TRUNCATION'):
    pass

class NullValueNoIndicatorParameter(DataError,
    code='22002', name='NULL_VALUE_NO_INDICATOR_PARAMETER'):
    pass

class NumericValueOutOfRange(DataError,
    code='22003', name='NUMERIC_VALUE_OUT_OF_RANGE'):
    pass

class NullValueNotAllowed(DataError,
    code='22004', name='NULL_VALUE_NOT_ALLOWED'):
    pass

class ErrorInAssignment(DataError,
    code='22005', name='ERROR_IN_ASSIGNMENT'):
    pass

class InvalidDatetimeFormat(DataError,
    code='22007', name='INVALID_DATETIME_FORMAT'):
    pass

class DatetimeFieldOverflow(DataError,
    code='22008', name='DATETIME_FIELD_OVERFLOW'):
    pass

class InvalidTimeZoneDisplacementValue(DataError,
    code='22009', name='INVALID_TIME_ZONE_DISPLACEMENT_VALUE'):
    pass

class EscapeCharacterConflict(DataError,
    code='2200B', name='ESCAPE_CHARACTER_CONFLICT'):
    pass

class InvalidUseOfEscapeCharacter(DataError,
    code='2200C', name='INVALID_USE_OF_ESCAPE_CHARACTER'):
    pass

class InvalidEscapeOctet(DataError,
    code='2200D', name='INVALID_ESCAPE_OCTET'):
    pass

class ZeroLengthCharacterString(DataError,
    code='2200F', name='ZERO_LENGTH_CHARACTER_STRING'):
    pass

class MostSpecificTypeMismatch(DataError,
    code='2200G', name='MOST_SPECIFIC_TYPE_MISMATCH'):
    pass

class SequenceGeneratorLimitExceeded(DataError,
    code='2200H', name='SEQUENCE_GENERATOR_LIMIT_EXCEEDED'):
    pass

class NotAnXmlDocument(DataError,
    code='2200L', name='NOT_AN_XML_DOCUMENT'):
    pass

class InvalidXmlDocument(DataError,
    code='2200M', name='INVALID_XML_DOCUMENT'):
    pass

class InvalidXmlContent(DataError,
    code='2200N', name='INVALID_XML_CONTENT'):
    pass

class InvalidXmlComment(DataError,
    code='2200S', name='INVALID_XML_COMMENT'):
    pass

class InvalidXmlProcessingInstruction(DataError,
    code='2200T', name='INVALID_XML_PROCESSING_INSTRUCTION'):
    pass

class InvalidIndicatorParameterValue(DataError,
    code='22010', name='INVALID_INDICATOR_PARAMETER_VALUE'):
    pass

class SubstringError(DataError,
    code='22011', name='SUBSTRING_ERROR'):
    pass

class DivisionByZero(DataError,
    code='22012', name='DIVISION_BY_ZERO'):
    pass

class InvalidPrecedingOrFollowingSize(DataError,
    code='22013', name='INVALID_PRECEDING_OR_FOLLOWING_SIZE'):
    pass

class InvalidArgumentForNtileFunction(DataError,
    code='22014', name='INVALID_ARGUMENT_FOR_NTILE_FUNCTION'):
    pass

class IntervalFieldOverflow(DataError,
    code='22015', name='INTERVAL_FIELD_OVERFLOW'):
    pass

class InvalidArgumentForNthValueFunction(DataError,
    code='22016', name='INVALID_ARGUMENT_FOR_NTH_VALUE_FUNCTION'):
    pass

class InvalidCharacterValueForCast(DataError,
    code='22018', name='INVALID_CHARACTER_VALUE_FOR_CAST'):
    pass

class InvalidEscapeCharacter(DataError,
    code='22019', name='INVALID_ESCAPE_CHARACTER'):
    pass

class InvalidRegularExpression(DataError,
    code='2201B', name='INVALID_REGULAR_EXPRESSION'):
    pass

class InvalidArgumentForLogarithm(DataError,
    code='2201E', name='INVALID_ARGUMENT_FOR_LOGARITHM'):
    pass

class InvalidArgumentForPowerFunction(DataError,
    code='2201F', name='INVALID_ARGUMENT_FOR_POWER_FUNCTION'):
    pass

class InvalidArgumentForWidthBucketFunction(DataError,
    code='2201G', name='INVALID_ARGUMENT_FOR_WIDTH_BUCKET_FUNCTION'):
    pass

class InvalidRowCountInLimitClause(DataError,
    code='2201W', name='INVALID_ROW_COUNT_IN_LIMIT_CLAUSE'):
    pass

class InvalidRowCountInResultOffsetClause(DataError,
    code='2201X', name='INVALID_ROW_COUNT_IN_RESULT_OFFSET_CLAUSE'):
    pass

class CharacterNotInRepertoire(DataError,
    code='22021', name='CHARACTER_NOT_IN_REPERTOIRE'):
    pass

class IndicatorOverflow(DataError,
    code='22022', name='INDICATOR_OVERFLOW'):
    pass

class InvalidParameterValue(DataError,
    code='22023', name='INVALID_PARAMETER_VALUE'):
    pass

class UnterminatedCString(DataError,
    code='22024', name='UNTERMINATED_C_STRING'):
    pass

class InvalidEscapeSequence(DataError,
    code='22025', name='INVALID_ESCAPE_SEQUENCE'):
    pass

class StringDataLengthMismatch(DataError,
    code='22026', name='STRING_DATA_LENGTH_MISMATCH'):
    pass

class TrimError(DataError,
    code='22027', name='TRIM_ERROR'):
    pass

class ArraySubscriptError(DataError,
    code='2202E', name='ARRAY_SUBSCRIPT_ERROR'):
    pass

class InvalidTablesampleRepeat(DataError,
    code='2202G', name='INVALID_TABLESAMPLE_REPEAT'):
    pass

class InvalidTablesampleArgument(DataError,
    code='2202H', name='INVALID_TABLESAMPLE_ARGUMENT'):
    pass

class DuplicateJsonObjectKeyValue(DataError,
    code='22030', name='DUPLICATE_JSON_OBJECT_KEY_VALUE'):
    pass

class InvalidArgumentForSqlJsonDatetimeFunction(DataError,
    code='22031', name='INVALID_ARGUMENT_FOR_SQL_JSON_DATETIME_FUNCTION'):
    pass

class InvalidJsonText(DataError,
    code='22032', name='INVALID_JSON_TEXT'):
    pass

class InvalidSqlJsonSubscript(DataError,
    code='22033', name='INVALID_SQL_JSON_SUBSCRIPT'):
    pass

class MoreThanOneSqlJsonItem(DataError,
    code='22034', name='MORE_THAN_ONE_SQL_JSON_ITEM'):
    pass

class NoSqlJsonItem(DataError,
    code='22035', name='NO_SQL_JSON_ITEM'):
    pass

class NonNumericSqlJsonItem(DataError,
    code='22036', name='NON_NUMERIC_SQL_JSON_ITEM'):
    pass

class NonUniqueKeysInAJsonObject(DataError,
    code='22037', name='NON_UNIQUE_KEYS_IN_A_JSON_OBJECT'):
    pass

class SingletonSqlJsonItemRequired(DataError,
    code='22038', name='SINGLETON_SQL_JSON_ITEM_REQUIRED'):
    pass

class SqlJsonArrayNotFound(DataError,
    code='22039', name='SQL_JSON_ARRAY_NOT_FOUND'):
    pass

class SqlJsonMemberNotFound(DataError,
    code='2203A', name='SQL_JSON_MEMBER_NOT_FOUND'):
    pass

class SqlJsonNumberNotFound(DataError,
    code='2203B', name='SQL_JSON_NUMBER_NOT_FOUND'):
    pass

class SqlJsonObjectNotFound(DataError,
    code='2203C', name='SQL_JSON_OBJECT_NOT_FOUND'):
    pass

class TooManyJsonArrayElements(DataError,
    code='2203D', name='TOO_MANY_JSON_ARRAY_ELEMENTS'):
    pass

class TooManyJsonObjectMembers(DataError,
    code='2203E', name='TOO_MANY_JSON_OBJECT_MEMBERS'):
    pass

class SqlJsonScalarRequired(DataError,
    code='2203F', name='SQL_JSON_SCALAR_REQUIRED'):
    pass

class SqlJsonItemCannotBeCastToTargetType(DataError,
    code='2203G', name='SQL_JSON_ITEM_CANNOT_BE_CAST_TO_TARGET_TYPE'):
    pass

class FloatingPointException(DataError,
    code='22P01', name='FLOATING_POINT_EXCEPTION'):
    pass

class InvalidTextRepresentation(DataError,
    code='22P02', name='INVALID_TEXT_REPRESENTATION'):
    pass

class InvalidBinaryRepresentation(DataError,
    code='22P03', name='INVALID_BINARY_REPRESENTATION'):
    pass

class BadCopyFileFormat(DataError,
    code='22P04', name='BAD_COPY_FILE_FORMAT'):
    pass

class UntranslatableCharacter(DataError,
    code='22P05', name='UNTRANSLATABLE_CHARACTER'):
    pass

class NonstandardUseOfEscapeCharacter(DataError,
    code='22P06', name='NONSTANDARD_USE_OF_ESCAPE_CHARACTER'):
    pass


# Class 23 - Integrity Constraint Violation

class IntegrityConstraintViolation(IntegrityError,
    code='23000', name='INTEGRITY_CONSTRAINT_VIOLATION'):
    pass

class RestrictViolation(IntegrityError,
    code='23001', name='RESTRICT_VIOLATION'):
    pass

class NotNullViolation(IntegrityError,
    code='23502', name='NOT_NULL_VIOLATION'):
    pass

class ForeignKeyViolation(IntegrityError,
    code='23503', name='FOREIGN_KEY_VIOLATION'):
    pass

class UniqueViolation(IntegrityError,
    code='23505', name='UNIQUE_VIOLATION'):
    pass

class CheckViolation(IntegrityError,
    code='23514', name='CHECK_VIOLATION'):
    pass

class ExclusionViolation(IntegrityError,
    code='23P01', name='EXCLUSION_VIOLATION'):
    pass


# Class 24 - Invalid Cursor State

class InvalidCursorState(InternalError,
    code='24000', name='INVALID_CURSOR_STATE'):
    pass


# Class 25 - Invalid Transaction State

class InvalidTransactionState(InternalError,
    code='25000', name='INVALID_TRANSACTION_STATE'):
    pass

class ActiveSqlTransaction(InternalError,
    code='25001', name='ACTIVE_SQL_TRANSACTION'):
    pass

class BranchTransactionAlreadyActive(InternalError,
    code='25002', name='BRANCH_TRANSACTION_ALREADY_ACTIVE'):
    pass

class InappropriateAccessModeForBranchTransaction(InternalError,
    code='25003', name='INAPPROPRIATE_ACCESS_MODE_FOR_BRANCH_TRANSACTION'):
    pass

class InappropriateIsolationLevelForBranchTransaction(InternalError,
    code='25004', name='INAPPROPRIATE_ISOLATION_LEVEL_FOR_BRANCH_TRANSACTION'):
    pass

class NoActiveSqlTransactionForBranchTransaction(InternalError,
    code='25005', name='NO_ACTIVE_SQL_TRANSACTION_FOR_BRANCH_TRANSACTION'):
    pass

class ReadOnlySqlTransaction(InternalError,
    code='25006', name='READ_ONLY_SQL_TRANSACTION'):
    pass

class SchemaAndDataStatementMixingNotSupported(InternalError,
    code='25007', name='SCHEMA_AND_DATA_STATEMENT_MIXING_NOT_SUPPORTED'):
    pass

class HeldCursorRequiresSameIsolationLevel(InternalError,
    code='25008', name='HELD_CURSOR_REQUIRES_SAME_ISOLATION_LEVEL'):
    pass

class NoActiveSqlTransaction(InternalError,
    code='25P01', name='NO_ACTIVE_SQL_TRANSACTION'):
    pass

class InFailedSqlTransaction(InternalError,
    code='25P02', name='IN_FAILED_SQL_TRANSACTION'):
    pass

class IdleInTransactionSessionTimeout(InternalError,
    code='25P03', name='IDLE_IN_TRANSACTION_SESSION_TIMEOUT'):
    pass

class TransactionTimeout(InternalError,
    code='25P04', name='TRANSACTION_TIMEOUT'):
    pass


# Class 26 - Invalid SQL Statement Name

class InvalidSqlStatementName(ProgrammingError,
    code='26000', name='INVALID_SQL_STATEMENT_NAME'):
    pass


# Class 27 - Triggered Data Change Violation

class TriggeredDataChangeViolation(OperationalError,
    code='27000', name='TRIGGERED_DATA_CHANGE_VIOLATION'):
    pass


# Class 28 - Invalid Authorization Specification

class InvalidAuthorizationSpecification(OperationalError,
    code='28000', name='INVALID_AUTHORIZATION_SPECIFICATION'):
    pass

class InvalidPassword(OperationalError,
    code='28P01', name='INVALID_PASSWORD'):
    pass


# Class 2B - Dependent Privilege Descriptors Still Exist

class DependentPrivilegeDescriptorsStillExist(InternalError,
    code='2B000', name='DEPENDENT_PRIVILEGE_DESCRIPTORS_STILL_EXIST'):
    pass

class DependentObjectsStillExist(InternalError,
    code='2BP01', name='DEPENDENT_OBJECTS_STILL_EXIST'):
    pass


# Class 2D - Invalid Transaction Termination

class InvalidTransactionTermination(InternalError,
    code='2D000', name='INVALID_TRANSACTION_TERMINATION'):
    pass


# Class 2F - SQL Routine Exception

class SqlRoutineException(OperationalError,
    code='2F000', name='SQL_ROUTINE_EXCEPTION'):
    pass

class ModifyingSqlDataNotPermitted(OperationalError,
    code='2F002', name='MODIFYING_SQL_DATA_NOT_PERMITTED'):
    pass

class ProhibitedSqlStatementAttempted(OperationalError,
    code='2F003', name='PROHIBITED_SQL_STATEMENT_ATTEMPTED'):
    pass

class ReadingSqlDataNotPermitted(OperationalError,
    code='2F004', name='READING_SQL_DATA_NOT_PERMITTED'):
    pass

class FunctionExecutedNoReturnStatement(OperationalError,
    code='2F005', name='FUNCTION_EXECUTED_NO_RETURN_STATEMENT'):
    pass


# Class 34 - Invalid Cursor Name

class InvalidCursorName(ProgrammingError,
    code='34000', name='INVALID_CURSOR_NAME'):
    pass


# Class 38 - External Routine Exception

class ExternalRoutineException(OperationalError,
    code='38000', name='EXTERNAL_ROUTINE_EXCEPTION'):
    pass

class ContainingSqlNotPermitted(OperationalError,
    code='38001', name='CONTAINING_SQL_NOT_PERMITTED'):
    pass

class ModifyingSqlDataNotPermittedExt(OperationalError,
    code='38002', name='MODIFYING_SQL_DATA_NOT_PERMITTED'):
    pass

class ProhibitedSqlStatementAttemptedExt(OperationalError,
    code='38003', name='PROHIBITED_SQL_STATEMENT_ATTEMPTED'):
    pass

class ReadingSqlDataNotPermittedExt(OperationalError,
    code='38004', name='READING_SQL_DATA_NOT_PERMITTED'):
    pass


# Class 39 - External Routine Invocation Exception

class ExternalRoutineInvocationException(OperationalError,
    code='39000', name='EXTERNAL_ROUTINE_INVOCATION_EXCEPTION'):
    pass

class InvalidSqlstateReturned(OperationalError,
    code='39001', name='INVALID_SQLSTATE_RETURNED'):
    pass

class NullValueNotAllowedExt(OperationalError,
    code='39004', name='NULL_VALUE_NOT_ALLOWED'):
    pass

class TriggerProtocolViolated(OperationalError,
    code='39P01', name='TRIGGER_PROTOCOL_VIOLATED'):
    pass

class SrfProtocolViolated(OperationalError,
    code='39P02', name='SRF_PROTOCOL_VIOLATED'):
    pass

class EventTriggerProtocolViolated(OperationalError,
    code='39P03', name='EVENT_TRIGGER_PROTOCOL_VIOLATED'):
    pass


# Class 3B - Savepoint Exception

class SavepointException(OperationalError,
    code='3B000', name='SAVEPOINT_EXCEPTION'):
    pass

class InvalidSavepointSpecification(OperationalError,
    code='3B001', name='INVALID_SAVEPOINT_SPECIFICATION'):
    pass


# Class 3D - Invalid Catalog Name

class InvalidCatalogName(ProgrammingError,
    code='3D000', name='INVALID_CATALOG_NAME'):
    pass


# Class 3F - Invalid Schema Name

class InvalidSchemaName(ProgrammingError,
    code='3F000', name='INVALID_SCHEMA_NAME'):
    pass


# Class 40 - Transaction Rollback

class TransactionRollback(OperationalError,
    code='40000', name='TRANSACTION_ROLLBACK'):
    pass

class SerializationFailure(OperationalError,
    code='40001', name='SERIALIZATION_FAILURE'):
    pass

class TransactionIntegrityConstraintViolation(OperationalError,
    code='40002', name='TRANSACTION_INTEGRITY_CONSTRAINT_VIOLATION'):
    pass

class StatementCompletionUnknown(OperationalError,
    code='40003', name='STATEMENT_COMPLETION_UNKNOWN'):
    pass

class DeadlockDetected(OperationalError,
    code='40P01', name='DEADLOCK_DETECTED'):
    pass


# Class 42 - Syntax Error or Access Rule Violation

class SyntaxErrorOrAccessRuleViolation(ProgrammingError,
    code='42000', name='SYNTAX_ERROR_OR_ACCESS_RULE_VIOLATION'):
    pass

class InsufficientPrivilege(ProgrammingError,
    code='42501', name='INSUFFICIENT_PRIVILEGE'):
    pass

class SyntaxError(ProgrammingError,
    code='42601', name='SYNTAX_ERROR'):
    pass

class InvalidName(ProgrammingError,
    code='42602', name='INVALID_NAME'):
    pass

class InvalidColumnDefinition(ProgrammingError,
    code='42611', name='INVALID_COLUMN_DEFINITION'):
    pass

class NameTooLong(ProgrammingError,
    code='42622', name='NAME_TOO_LONG'):
    pass

class DuplicateColumn(ProgrammingError,
    code='42701', name='DUPLICATE_COLUMN'):
    pass

class AmbiguousColumn(ProgrammingError,
    code='42702', name='AMBIGUOUS_COLUMN'):
    pass

class UndefinedColumn(ProgrammingError,
    code='42703', name='UNDEFINED_COLUMN'):
    pass

class UndefinedObject(ProgrammingError,
    code='42704', name='UNDEFINED_OBJECT'):
    pass

class DuplicateObject(ProgrammingError,
    code='42710', name='DUPLICATE_OBJECT'):
    pass

class DuplicateAlias(ProgrammingError,
    code='42712', name='DUPLICATE_ALIAS'):
    pass

class DuplicateFunction(ProgrammingError,
    code='42723', name='DUPLICATE_FUNCTION'):
    pass

class AmbiguousFunction(ProgrammingError,
    code='42725', name='AMBIGUOUS_FUNCTION'):
    pass

class GroupingError(ProgrammingError,
    code='42803', name='GROUPING_ERROR'):
    pass

class DatatypeMismatch(ProgrammingError,
    code='42804', name='DATATYPE_MISMATCH'):
    pass

class WrongObjectType(ProgrammingError,
    code='42809', name='WRONG_OBJECT_TYPE'):
    pass

class InvalidForeignKey(ProgrammingError,
    code='42830', name='INVALID_FOREIGN_KEY'):
    pass

class CannotCoerce(ProgrammingError,
    code='42846', name='CANNOT_COERCE'):
    pass

class UndefinedFunction(ProgrammingError,
    code='42883', name='UNDEFINED_FUNCTION'):
    pass

class GeneratedAlways(ProgrammingError,
    code='428C9', name='GENERATED_ALWAYS'):
    pass

class ReservedName(ProgrammingError,
    code='42939', name='RESERVED_NAME'):
    pass

class UndefinedTable(ProgrammingError,
    code='42P01', name='UNDEFINED_TABLE'):
    pass

class UndefinedParameter(ProgrammingError,
    code='42P02', name='UNDEFINED_PARAMETER'):
    pass

class DuplicateCursor(ProgrammingError,
    code='42P03', name='DUPLICATE_CURSOR'):
    pass

class DuplicateDatabase(ProgrammingError,
    code='42P04', name='DUPLICATE_DATABASE'):
    pass

class DuplicatePreparedStatement(ProgrammingError,
    code='42P05', name='DUPLICATE_PREPARED_STATEMENT'):
    pass

class DuplicateSchema(ProgrammingError,
    code='42P06', name='DUPLICATE_SCHEMA'):
    pass

class DuplicateTable(ProgrammingError,
    code='42P07', name='DUPLICATE_TABLE'):
    pass

class AmbiguousParameter(ProgrammingError,
    code='42P08', name='AMBIGUOUS_PARAMETER'):
    pass

class AmbiguousAlias(ProgrammingError,
    code='42P09', name='AMBIGUOUS_ALIAS'):
    pass

class InvalidColumnReference(ProgrammingError,
    code='42P10', name='INVALID_COLUMN_REFERENCE'):
    pass

class InvalidCursorDefinition(ProgrammingError,
    code='42P11', name='INVALID_CURSOR_DEFINITION'):
    pass

class InvalidDatabaseDefinition(ProgrammingError,
    code='42P12', name='INVALID_DATABASE_DEFINITION'):
    pass

class InvalidFunctionDefinition(ProgrammingError,
    code='42P13', name='INVALID_FUNCTION_DEFINITION'):
    pass

class InvalidPreparedStatementDefinition(ProgrammingError,
    code='42P14', name='INVALID_PREPARED_STATEMENT_DEFINITION'):
    pass

class InvalidSchemaDefinition(ProgrammingError,
    code='42P15', name='INVALID_SCHEMA_DEFINITION'):
    pass

class InvalidTableDefinition(ProgrammingError,
    code='42P16', name='INVALID_TABLE_DEFINITION'):
    pass

class InvalidObjectDefinition(ProgrammingError,
    code='42P17', name='INVALID_OBJECT_DEFINITION'):
    pass

class IndeterminateDatatype(ProgrammingError,
    code='42P18', name='INDETERMINATE_DATATYPE'):
    pass

class InvalidRecursion(ProgrammingError,
    code='42P19', name='INVALID_RECURSION'):
    pass

class WindowingError(ProgrammingError,
    code='42P20', name='WINDOWING_ERROR'):
    pass

class CollationMismatch(ProgrammingError,
    code='42P21', name='COLLATION_MISMATCH'):
    pass

class IndeterminateCollation(ProgrammingError,
    code='42P22', name='INDETERMINATE_COLLATION'):
    pass


# Class 44 - WITH CHECK OPTION Violation

class WithCheckOptionViolation(ProgrammingError,
    code='44000', name='WITH_CHECK_OPTION_VIOLATION'):
    pass


# Class 53 - Insufficient Resources

class InsufficientResources(OperationalError,
    code='53000', name='INSUFFICIENT_RESOURCES'):
    pass

class DiskFull(OperationalError,
    code='53100', name='DISK_FULL'):
    pass

class OutOfMemory(OperationalError,
    code='53200', name='OUT_OF_MEMORY'):
    pass

class TooManyConnections(OperationalError,
    code='53300', name='TOO_MANY_CONNECTIONS'):
    pass

class ConfigurationLimitExceeded(OperationalError,
    code='53400', name='CONFIGURATION_LIMIT_EXCEEDED'):
    pass


# Class 54 - Program Limit Exceeded

class ProgramLimitExceeded(OperationalError,
    code='54000', name='PROGRAM_LIMIT_EXCEEDED'):
    pass

class StatementTooComplex(OperationalError,
    code='54001', name='STATEMENT_TOO_COMPLEX'):
    pass

class TooManyColumns(OperationalError,
    code='54011', name='TOO_MANY_COLUMNS'):
    pass

class TooManyArguments(OperationalError,
    code='54023', name='TOO_MANY_ARGUMENTS'):
    pass


# Class 55 - Object Not In Prerequisite State

class ObjectNotInPrerequisiteState(OperationalError,
    code='55000', name='OBJECT_NOT_IN_PREREQUISITE_STATE'):
    pass

class ObjectInUse(OperationalError,
    code='55006', name='OBJECT_IN_USE'):
    pass

class CantChangeRuntimeParam(OperationalError,
    code='55P02', name='CANT_CHANGE_RUNTIME_PARAM'):
    pass

class LockNotAvailable(OperationalError,
    code='55P03', name='LOCK_NOT_AVAILABLE'):
    pass

class UnsafeNewEnumValueUsage(OperationalError,
    code='55P04', name='UNSAFE_NEW_ENUM_VALUE_USAGE'):
    pass


# Class 57 - Operator Intervention

class OperatorIntervention(OperationalError,
    code='57000', name='OPERATOR_INTERVENTION'):
    pass

class QueryCanceled(OperationalError,
    code='57014', name='QUERY_CANCELED'):
    pass

class AdminShutdown(OperationalError,
    code='57P01', name='ADMIN_SHUTDOWN'):
    pass

class CrashShutdown(OperationalError,
    code='57P02', name='CRASH_SHUTDOWN'):
    pass

class CannotConnectNow(OperationalError,
    code='57P03', name='CANNOT_CONNECT_NOW'):
    pass

class DatabaseDropped(OperationalError,
    code='57P04', name='DATABASE_DROPPED'):
    pass

class IdleSessionTimeout(OperationalError,
    code='57P05', name='IDLE_SESSION_TIMEOUT'):
    pass


# Class 58 - System Error (errors external to PostgreSQL itself)

class SystemError(OperationalError,
    code='58000', name='SYSTEM_ERROR'):
    pass

class IoError(OperationalError,
    code='58030', name='IO_ERROR'):
    pass

class UndefinedFile(OperationalError,
    code='58P01', name='UNDEFINED_FILE'):
    pass

class DuplicateFile(OperationalError,
    code='58P02', name='DUPLICATE_FILE'):
    pass

class FileNameTooLong(OperationalError,
    code='58P03', name='FILE_NAME_TOO_LONG'):
    pass


# Class 72 - Snapshot Failure

class SnapshotTooOld(DatabaseError,
    code='72000', name='SNAPSHOT_TOO_OLD'):
    pass


# Class F0 - Configuration File Error

class ConfigFileError(OperationalError,
    code='F0000', name='CONFIG_FILE_ERROR'):
    pass

class LockFileExists(OperationalError,
    code='F0001', name='LOCK_FILE_EXISTS'):
    pass


# Class HV - Foreign Data Wrapper Error (SQL/MED)

class FdwError(OperationalError,
    code='HV000', name='FDW_ERROR'):
    pass

class FdwOutOfMemory(OperationalError,
    code='HV001', name='FDW_OUT_OF_MEMORY'):
    pass

class FdwDynamicParameterValueNeeded(OperationalError,
    code='HV002', name='FDW_DYNAMIC_PARAMETER_VALUE_NEEDED'):
    pass

class FdwInvalidDataType(OperationalError,
    code='HV004', name='FDW_INVALID_DATA_TYPE'):
    pass

class FdwColumnNameNotFound(OperationalError,
    code='HV005', name='FDW_COLUMN_NAME_NOT_FOUND'):
    pass

class FdwInvalidDataTypeDescriptors(OperationalError,
    code='HV006', name='FDW_INVALID_DATA_TYPE_DESCRIPTORS'):
    pass

class FdwInvalidColumnName(OperationalError,
    code='HV007', name='FDW_INVALID_COLUMN_NAME'):
    pass

class FdwInvalidColumnNumber(OperationalError,
    code='HV008', name='FDW_INVALID_COLUMN_NUMBER'):
    pass

class FdwInvalidUseOfNullPointer(OperationalError,
    code='HV009', name='FDW_INVALID_USE_OF_NULL_POINTER'):
    pass

class FdwInvalidStringFormat(OperationalError,
    code='HV00A', name='FDW_INVALID_STRING_FORMAT'):
    pass

class FdwInvalidHandle(OperationalError,
    code='HV00B', name='FDW_INVALID_HANDLE'):
    pass

class FdwInvalidOptionIndex(OperationalError,
    code='HV00C', name='FDW_INVALID_OPTION_INDEX'):
    pass

class FdwInvalidOptionName(OperationalError,
    code='HV00D', name='FDW_INVALID_OPTION_NAME'):
    pass

class FdwOptionNameNotFound(OperationalError,
    code='HV00J', name='FDW_OPTION_NAME_NOT_FOUND'):
    pass

class FdwReplyHandle(OperationalError,
    code='HV00K', name='FDW_REPLY_HANDLE'):
    pass

class FdwUnableToCreateExecution(OperationalError,
    code='HV00L', name='FDW_UNABLE_TO_CREATE_EXECUTION'):
    pass

class FdwUnableToCreateReply(OperationalError,
    code='HV00M', name='FDW_UNABLE_TO_CREATE_REPLY'):
    pass

class FdwUnableToEstablishConnection(OperationalError,
    code='HV00N', name='FDW_UNABLE_TO_ESTABLISH_CONNECTION'):
    pass

class FdwNoSchemas(OperationalError,
    code='HV00P', name='FDW_NO_SCHEMAS'):
    pass

class FdwSchemaNotFound(OperationalError,
    code='HV00Q', name='FDW_SCHEMA_NOT_FOUND'):
    pass

class FdwTableNotFound(OperationalError,
    code='HV00R', name='FDW_TABLE_NOT_FOUND'):
    pass

class FdwFunctionSequenceError(OperationalError,
    code='HV010', name='FDW_FUNCTION_SEQUENCE_ERROR'):
    pass

class FdwTooManyHandles(OperationalError,
    code='HV014', name='FDW_TOO_MANY_HANDLES'):
    pass

class FdwInconsistentDescriptorInformation(OperationalError,
    code='HV021', name='FDW_INCONSISTENT_DESCRIPTOR_INFORMATION'):
    pass

class FdwInvalidAttributeValue(OperationalError,
    code='HV024', name='FDW_INVALID_ATTRIBUTE_VALUE'):
    pass

class FdwInvalidStringLengthOrBufferLength(OperationalError,
    code='HV090', name='FDW_INVALID_STRING_LENGTH_OR_BUFFER_LENGTH'):
    pass

class FdwInvalidDescriptorFieldIdentifier(OperationalError,
    code='HV091', name='FDW_INVALID_DESCRIPTOR_FIELD_IDENTIFIER'):
    pass


# Class P0 - PL/pgSQL Error

class PlpgsqlError(ProgrammingError,
    code='P0000', name='PLPGSQL_ERROR'):
    pass

class RaiseException(ProgrammingError,
    code='P0001', name='RAISE_EXCEPTION'):
    pass

class NoDataFound(ProgrammingError,
    code='P0002', name='NO_DATA_FOUND'):
    pass

class TooManyRows(ProgrammingError,
    code='P0003', name='TOO_MANY_ROWS'):
    pass

class AssertFailure(ProgrammingError,
    code='P0004', name='ASSERT_FAILURE'):
    pass


# Class XX - Internal Error

class InternalError_(InternalError,
    code='XX000', name='INTERNAL_ERROR'):
    pass

class DataCorrupted(InternalError,
    code='XX001', name='DATA_CORRUPTED'):
    pass

class IndexCorrupted(InternalError,
    code='XX002', name='INDEX_CORRUPTED'):
    pass


# autogenerated: end
# fmt: on

# Don't show a complete traceback upon raising these exception.
# Usually the traceback starts from internal functions (for instance in the
# server communication callbacks) but, for the end user, it's more important
# to get the high level information about where the exception was raised, for
# instance in a certain `Cursor.execute()`.

_NO_TRACEBACK = (Error, KeyboardInterrupt, CancelledError)



"""
psycopg_c.pq.PGresult object implementation.
"""

# Copyright (C) 2020 The Psycopg Team

cimport cython
from cpython.mem cimport PyMem_Free, PyMem_Malloc

from psycopg.pq.misc import PGresAttDesc
from psycopg.pq._enums import ExecStatus


@cython.freelist(8)
cdef class PGresult:
    def __cinit__(self):
        self._pgresult_ptr = NULL

    @staticmethod
    cdef PGresult _from_ptr(libpq.PGresult *ptr):
        cdef PGresult rv = PGresult.__new__(PGresult)
        rv._pgresult_ptr = ptr
        return rv

    def __dealloc__(self) -> None:
        self.clear()

    def __repr__(self) -> str:
        cls = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
        status = ExecStatus(self.status)
        return f"<{cls} [{status.name}] at 0x{id(self):x}>"

    def clear(self) -> None:
        if self._pgresult_ptr is not NULL:
            libpq.PQclear(self._pgresult_ptr)
            self._pgresult_ptr = NULL

    @property
    def pgresult_ptr(self) -> int | None:
        if self._pgresult_ptr:
            return <long long><void *>self._pgresult_ptr
        else:
            return None

    @property
    def status(self) -> int:
        return libpq.PQresultStatus(self._pgresult_ptr)

    @property
    def error_message(self) -> bytes:
        return libpq.PQresultErrorMessage(self._pgresult_ptr)

    def get_error_message(self, encoding: str = "utf-8") -> str:
        return _clean_error_message(self.error_message, encoding)

    def error_field(self, int fieldcode) -> bytes | None:
        cdef char * rv = libpq.PQresultErrorField(self._pgresult_ptr, fieldcode)
        if rv is not NULL:
            return rv
        else:
            return None

    @property
    def ntuples(self) -> int:
        return libpq.PQntuples(self._pgresult_ptr)

    @property
    def nfields(self) -> int:
        return libpq.PQnfields(self._pgresult_ptr)

    def fname(self, int column_number) -> bytes | None:
        cdef char *rv = libpq.PQfname(self._pgresult_ptr, column_number)
        if rv is not NULL:
            return rv
        else:
            return None

    def ftable(self, int column_number) -> int:
        return libpq.PQftable(self._pgresult_ptr, column_number)

    def ftablecol(self, int column_number) -> int:
        return libpq.PQftablecol(self._pgresult_ptr, column_number)

    def fformat(self, int column_number) -> int:
        return libpq.PQfformat(self._pgresult_ptr, column_number)

    def ftype(self, int column_number) -> int:
        return libpq.PQftype(self._pgresult_ptr, column_number)

    def fmod(self, int column_number) -> int:
        return libpq.PQfmod(self._pgresult_ptr, column_number)

    def fsize(self, int column_number) -> int:
        return libpq.PQfsize(self._pgresult_ptr, column_number)

    @property
    def binary_tuples(self) -> int:
        return libpq.PQbinaryTuples(self._pgresult_ptr)

    def get_value(self, int row_number, int column_number) -> bytes | None:
        cdef int crow = row_number
        cdef int ccol = column_number
        cdef int length = libpq.PQgetlength(self._pgresult_ptr, crow, ccol)
        cdef char *v
        if length:
            v = libpq.PQgetvalue(self._pgresult_ptr, crow, ccol)
            # TODO: avoid copy
            return v[:length]
        else:
            if libpq.PQgetisnull(self._pgresult_ptr, crow, ccol):
                return None
            else:
                return b""

    @property
    def nparams(self) -> int:
        return libpq.PQnparams(self._pgresult_ptr)

    def param_type(self, int param_number) -> int:
        return libpq.PQparamtype(self._pgresult_ptr, param_number)

    @property
    def command_status(self) -> bytes | None:
        cdef char *rv = libpq.PQcmdStatus(self._pgresult_ptr)
        if rv is not NULL:
            return rv
        else:
            return None

    @property
    def command_tuples(self) -> int | None:
        cdef char *rv = libpq.PQcmdTuples(self._pgresult_ptr)
        if rv is NULL:
            return None
        cdef bytes brv = rv
        return int(brv) if brv else None

    @property
    def oid_value(self) -> int:
        return libpq.PQoidValue(self._pgresult_ptr)

    def set_attributes(self, descriptions: list[PGresAttDesc]):
        cdef Py_ssize_t num = len(descriptions)
        cdef libpq.PGresAttDesc *attrs = <libpq.PGresAttDesc *>PyMem_Malloc(
            num * sizeof(libpq.PGresAttDesc))

        for i in range(num):
            descr = descriptions[i]
            attrs[i].name = descr.name
            attrs[i].tableid = descr.tableid
            attrs[i].columnid = descr.columnid
            attrs[i].format = descr.format
            attrs[i].typid = descr.typid
            attrs[i].typlen = descr.typlen
            attrs[i].atttypmod = descr.atttypmod

        cdef int res = libpq.PQsetResultAttrs(self._pgresult_ptr, <int>num, attrs)
        PyMem_Free(attrs)
        if (res == 0):
            raise e.OperationalError("PQsetResultAttrs failed")
